import time
from random import randint

from test_junkie.decorators import Suite, test, beforeClass, beforeTest, afterTest, afterClass
from test_junkie.meta import meta, Meta
from tests.junkie_suites.TestListener import TestListener
from tests.junkie_suites.TestRules import TestRules


@Suite(retry=2,
       listener=TestListener,
       priority=2, feature="Store", owner="Mike")
class NewProductsSuite:

    @beforeClass()
    def before_class1(self):
        pass

    @beforeTest()
    def before_test2(self):
        # write your code here
        pass

    @afterTest()
    def after_test3(self):
        # write your code here
        pass

    @afterClass()
    def after_class4(self):
        # write your code here
        pass

    @test(component="Admin", tags=["store_management"])
    def add_new_product(self):
        # time.sleep(randint(0, 4))
        pass

    @test(component="Admin", tags=["store_management"])
    def remove_product(self):
        # time.sleep(randint(0, 4))
        pass

    @test(component="Admin", tags=["store_management"])
    def publish_product(self):
        # time.sleep(randint(0, 4))
        pass

    @test(component="Admin", tags=["store_management"])
    def edit_product(self):
        # time.sleep(randint(0, 4))
        pass
        raise Exception("Product not found")

    @test(component="Admin", tags=["store_management"])
    def archive_product(self):
        # time.sleep(randint(0, 4))
        pass
        raise Exception("Product not found")



import errno
import multiprocessing
import os
import sys
import threading
import time
import traceback
from datetime import datetime

import pkg_resources

from test_junkie.cli.cli import CliUtils

from test_junkie.cli.cli_config import Config
from test_junkie.debugger import LogJunkie
from test_junkie.decorators import DecoratorType
from test_junkie.constants import SuiteCategory, TestCategory, DocumentationLinks


class ClassMetrics(object):

    def __init__(self):

        self.__stats = {"status": None, "retry": 0, "runtime": 0, "start": None, "end": None,
                        DecoratorType.BEFORE_CLASS: {"performance": [], "exceptions": [], "tracebacks": []},
                        DecoratorType.BEFORE_TEST: {"performance": [], "exceptions": [], "tracebacks": []},
                        DecoratorType.AFTER_TEST: {"performance": [], "exceptions": [], "tracebacks": []},
                        DecoratorType.AFTER_CLASS: {"performance": [], "exceptions": [], "tracebacks": []}}

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    def update_decorator_metrics(self, decorator, start_time, exception=None, trace=None):
        from test_junkie.objects import Limiter
        self.__stats[decorator]["performance"].append(time.time() - start_time)
        self.__stats[decorator]["exceptions"].append(Limiter.parse_exception_object(exception))
        self.__stats[decorator]["tracebacks"].append(Limiter.parse_traceback(trace))

    def update_suite_metrics(self, status, start_time, initiation_error=None):

        self.__stats["status"] = status
        if status not in [SuiteCategory.CANCEL, SuiteCategory.SKIP, SuiteCategory.IGNORE]:
            self.__stats["retry"] += 1
        self.__stats["start"] = start_time
        self.__stats["end"] = time.time()
        self.__stats["runtime"] = self.__stats["end"] - self.__stats["start"]
        if status in [SuiteCategory.IGNORE] and initiation_error is not None:
            self.__stats.update({"initiation_error": initiation_error})

    def get_metrics(self):

        return self.__stats


class TestMetrics(object):

    def __init__(self):

        self.__stats = {}

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    def update_metrics(self, status, start_time, param=None, class_param=None, exception=None,
                       formatted_traceback=None, runtime=None, decorator=None):

        def __get_template():

            return {"status": None,
                    "retry": 0,
                    "performance": [],
                    "exceptions": [],
                    "tracebacks": [],
                    DecoratorType.BEFORE_TEST: {"performance": [], "exceptions": [], "tracebacks": []},
                    DecoratorType.AFTER_TEST: {"performance": [], "exceptions": [], "tracebacks": []}}

        runtime = runtime if runtime is not None else time.time() - start_time
        string_param = str(param)
        string_class_param = str(class_param)
        if string_class_param not in self.__stats:
            self.__stats.update({string_class_param: {string_param: __get_template()}})
        elif string_param not in self.__stats[string_class_param]:
            self.__stats[string_class_param].update({string_param: __get_template()})

        from test_junkie.objects import Limiter
        if decorator is not None:
            self.__stats[string_class_param][string_param][decorator]["performance"].append(time.time() - start_time)
            self.__stats[string_class_param][string_param][decorator]["exceptions"]\
                .append(Limiter.parse_exception_object(exception))
            self.__stats[string_class_param][string_param][decorator]["tracebacks"]\
                .append(Limiter.parse_traceback(formatted_traceback))

            if self.__stats[string_class_param][string_param]["status"] is None:
                # while status is not set, will use one that is currently passed in
                # Once its explicitly updated for the test decorator, then we wont change it
                self.__stats[string_class_param][string_param]["status"] = status
        else:
            self.__stats[string_class_param][string_param]["performance"].append(runtime)
            self.__stats[string_class_param][string_param]["exceptions"]\
                .append(Limiter.parse_exception_object(exception))
            self.__stats[string_class_param][string_param]["tracebacks"]\
                .append(Limiter.parse_traceback(formatted_traceback))
            self.__stats[string_class_param][string_param]["retry"] += 1
            self.__stats[string_class_param][string_param]["status"] = status
            self.__stats[string_class_param][string_param]["param"] = param
            self.__stats[string_class_param][string_param]["class_param"] = class_param

    def get_metrics(self):

        return self.__stats


class Aggregator(object):

    def __init__(self, executed_suites):

        self.__executed_suites = executed_suites

    @property
    def executed_suites(self):
        return self.__executed_suites

    @staticmethod
    def percentage(total, part):
        if part > 0:
            return "{:0.2f}".format(float(part) / float(total) * 100)
        else:
            return "0"

    def get_basic_report(self):

        def get_template():

            return {"total": 0,
                    TestCategory.SUCCESS: 0,
                    TestCategory.FAIL: 0,
                    TestCategory.ERROR: 0,
                    TestCategory.IGNORE: 0,
                    TestCategory.SKIP: 0,
                    TestCategory.CANCEL: 0}

        report = {"tests": get_template(),
                  "suites": {}}

        for suite in self.__executed_suites:
            if suite not in report["suites"]:
                report["suites"].update({suite: get_template()})
            for test in suite.get_test_objects():
                test_metrics = test.metrics.get_metrics()
                for class_param, class_param_data in test_metrics.items():
                    for param, param_data in class_param_data.items():
                        report["tests"]["total"] += 1
                        report["tests"][param_data["status"]] += 1
                        report["suites"][suite]["total"] += 1
                        report["suites"][suite][param_data["status"]] += 1
        return report

    def get_report_by_features(self):

        report = {}
        for suite in self.__executed_suites:
            feature = suite.get_feature()
            if feature not in report:
                report.update({feature: {"_totals_": Aggregator.get_template()}})
            for test in suite.get_test_objects():
                component = test.get_component()
                if component not in report[feature]:
                    report[feature].update({component: Aggregator.get_template()})
                test_metrics = test.metrics.get_metrics()
                Aggregator._update_report(report, test_metrics, feature, component)
        return report

    def get_report_by_tags(self):

        report = {}
        for suite in self.__executed_suites:
            tag_metrics = suite.get_data_by_tags()
            for metric, metric_data in tag_metrics.items():
                if metric == "_totals_":
                    continue
                metric = metric if metric is not None else "Not Defined"
                if metric not in report:
                    report.update({metric: Aggregator.get_template()})
                for entry in metric_data["performance"]:
                    report[metric]["performance"].append(entry)
                for entry in metric_data["exceptions"]:
                    if entry is not None:
                        report[metric]["exceptions"].append(entry)
                for retry in metric_data["retries"]:
                    report[metric]["retries"].append(retry)
                report[metric]["total"] += metric_data["total"]
                for status in TestCategory.ALL:
                    report[metric][status] += metric_data[status]
        return report

    def get_report_by_owner(self):

        report = {"_totals_": Aggregator.get_template()}
        for suite in self.__executed_suites:
            for test in suite.get_test_objects():
                owner = test.get_owner()
                if owner not in report:
                    report.update({owner: Aggregator.get_template()})
                test_metrics = test.metrics.get_metrics()
                Aggregator._update_report(report, test_metrics, owner)
        return report

    def get_report_by_suite(self):

        report = {"_totals_": Aggregator.get_template()}
        for suite in self.__executed_suites:
            suite_name = suite.get_class_name()
            for test in suite.get_test_objects():
                if suite_name not in report:
                    report.update({suite_name: Aggregator.get_template()})
                test_metrics = test.metrics.get_metrics()
                Aggregator._update_report(report, test_metrics, suite_name)
        return report

    @staticmethod
    def _update_report(report, metrics, category, subcategory=0):

        for class_param, class_param_data in metrics.items():
            for param, data in class_param_data.items():
                for entry in data["performance"]:
                    if subcategory == 0:
                        report[category]["performance"].append(entry)
                        report["_totals_"]["performance"].append(entry)
                    else:
                        report[category][subcategory]["performance"].append(entry)
                        report[category]["_totals_"]["performance"].append(entry)
                for entry in data["exceptions"]:
                    if entry is not None:
                        if subcategory == 0:
                            report[category]["exceptions"].append(entry)
                            report["_totals_"]["exceptions"].append(entry)
                        else:
                            report[category][subcategory]["exceptions"].append(entry)
                            report[category]["_totals_"]["exceptions"].append(entry)
                if subcategory == 0:
                    report[category]["retries"].append(data["retry"])
                    report["_totals_"]["retries"].append(data["retry"])
                    report[category]["total"] += 1
                    report["_totals_"]["total"] += 1
                    report[category][data["status"]] += 1
                    report["_totals_"][data["status"]] += 1
                else:
                    report[category][subcategory]["retries"].append(data["retry"])
                    report[category]["_totals_"]["retries"].append(data["retry"])
                    report[category][subcategory]["total"] += 1
                    report[category]["_totals_"]["total"] += 1
                    report[category][subcategory][data["status"]] += 1
                    report[category]["_totals_"][data["status"]] += 1

        return report

    @staticmethod
    def present_console_output(aggregator):

        def parse_exception(value):
            if value is not None:
                error = ""
                for line in value.split("\n"):
                    error += "\n\t\t  {}".format(line)
                return error

        report = aggregator.get_basic_report()
        test_report = report["tests"]
        suite_report = report["suites"]
        for status in TestCategory.ALL:
            value = "[{part}/{total} {percent}%] {status}".format(part=test_report[status],
                                                                  total=test_report["total"],
                                                                  status=status.upper(),
                                                                  percent=Aggregator.percentage(test_report["total"],
                                                                                                test_report[status]))
            if test_report[status]:
                value = CliUtils.format_bold_string(value)
            print(value)
        print("")
        for suite, stats in suite_report.items():
            status = suite.metrics.get_metrics()["status"]
            if status is None:  # this means that something went wrong with custom event processing
                status = "*"+SuiteCategory.ERROR
            print(">> [{status}] [{passed}/{total} {rate}%] [{runtime:0.2f}s] {module}.{name}"
                  .format(module=CliUtils.format_bold_string(suite.get_class_module()),
                          name=CliUtils.format_bold_string(suite.get_class_name()),
                          status=CliUtils.format_bold_string(status.upper()),
                          runtime=suite.get_runtime(),
                          rate=Aggregator.percentage(stats["total"], stats[TestCategory.SUCCESS]),
                          passed=stats[TestCategory.SUCCESS],
                          total=stats["total"]))
            if status == SuiteCategory.IGNORE:
                print("\t|__ reason: {error}".format(error=suite.metrics.get_metrics().get("initiation_error", None)))
            if status != SuiteCategory.SUCCESS:
                tests = suite.get_unsuccessful_tests()
                for test in tests:
                    test_metrics = test.metrics.get_metrics()
                    print("\t|__ test: {name}()".format(name=CliUtils.format_bold_string(test.get_function_name())))
                    for class_param, class_param_data in test_metrics.items():
                        if class_param != "None":
                            print("\t  |__ class parameter: {class_parameter}".format(class_parameter=class_param))
                        for param, param_data in class_param_data.items():
                            if param != "None":
                                print("\t    |__ parameter: {parameter}".format(parameter=param))
                            for index in range(param_data["retry"]):
                                trace = param_data["tracebacks"][index]
                                print("\t      |__ run #{num} [{status}] [{runtime:0.2f}s] :: Traceback: {exception}"
                                      .format(num=index + 1,
                                              exception=CliUtils.format_color_string(parse_exception(trace), "red"),
                                              runtime=param_data["performance"][index],
                                              status=param_data["status"].upper()))
        print("\n===========================================================")
        print(". Test Junkie {} (Python{}) {} .".format(pkg_resources.require("test-junkie")[0].version,
                                                        sys.version_info[0],
                                                        DocumentationLinks.DOMAIN))
        print("===========================================================")

    @staticmethod
    def get_template():

        return {"performance": [],
                "exceptions": [],
                "retries": [],
                "total": 0,
                TestCategory.SUCCESS: 0,
                TestCategory.SKIP: 0,
                TestCategory.FAIL: 0,
                TestCategory.CANCEL: 0,
                TestCategory.IGNORE: 0,
                TestCategory.ERROR: 0}

    def get_average_test_runtime(self):

        from statistics import mean
        samples = []
        for suite in self.__executed_suites:
            for test in suite.get_test_objects():
                test_metrics = test.metrics.get_metrics()
                for class_param, class_param_data in test_metrics.items():
                    for param, param_data in class_param_data.items():
                        samples.append(mean(param_data["performance"]))
        return mean(samples) if samples else None


class ResourceMonitor(threading.Thread):

    def __init__(self):

        self.file_path = "{dir}{sep}.resources_{timestamp}".format(dir=Config.get_root_dir(),
                                                                   sep=os.sep,
                                                                   timestamp=time.time())

        threading.Thread.__init__(self)
        self.exit = multiprocessing.Event()

    def get_file_path(self):

        return self.file_path

    def mkdir_p(self, path):
        try:
            os.makedirs(path)
        except OSError as exc:  # Python >2.5
            if exc.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else:
                raise

    def run(self):
        import psutil
        self.mkdir_p(Config.get_root_dir())
        with open(self.file_path, "w+") as records:
            records.write("")
        while not self.exit.is_set():
            time.sleep(1)
            data = "{timestamp}, {cpu}, {memory}\n".format(timestamp=datetime.now(),
                                                           cpu=psutil.cpu_percent(),
                                                           memory=psutil.virtual_memory().percent)
            with open(self.file_path, "a+") as records:
                records.write(data)

    def shutdown(self):
        self.exit.set()

    def cleanup(self):
        try:
            os.remove(self.file_path)
        except:
            LogJunkie.error(traceback.format_exc())



from test_junkie.decorators import test, beforeTest, afterTest, afterClass, beforeClass, Suite


def skip_function(meta):
    assert meta.get("name") == "skip_function"
    return True


@Suite()
class BasicSuite:

    @beforeClass()
    def before_class(self):
        pass

    @beforeTest()
    def before_test(self):
        pass

    @afterTest()
    def after_test(self):
        pass

    @afterClass()
    def after_class(self):
        pass

    @test()
    def failure(self):
        assert True is False

    @test()
    def error(self):
        raise Exception("Exception")

    @test(skip=True)
    def skip(self):
        pass

    @test(meta={"name": "skip_function"},
          skip=skip_function)
    def skip_function(self):
        pass

    @test(parameters=[1, 2, 3, 4])
    def parameters(self, parameter):
        pass

    @test(retry=2)
    def retry(self):
        assert True is False



import ast
from test_junkie.constants import DocumentationLinks, Undefined
from test_junkie.debugger import LogJunkie
from test_junkie.errors import BadParameters
from test_junkie.cli.cli_config import Config


class Settings:

    __DEFAULT_TEST_THREAD_LIMIT = 1
    __DEFAULT_SUITE_THREAD_LIMIT = 1
    __DEFAULT_COMPONENTS = None
    __DEFAULT_FEATURES = None
    __DEFAULT_OWNERS = None
    __DEFAULT_TAGS = None
    __DEFAULT_HTML = None
    __DEFAULT_XML = None
    __DEFAULT_TESTS = None
    __DEFAULT_RESOURCE_MON = False
    __DEFAULT_QUIET = False

    def __init__(self, runner_kwargs, run_kwargs):
        """

        :param runner_kwargs: DICT, arguments that are passed in to initiate the Runner instance
        :param run_kwargs: DICT, arguments that are passes to the run() method of the Runner instance
        """
        runner_kwargs.update(run_kwargs)

        self.kwargs = runner_kwargs

        self.config = None
        if self.kwargs.get("config", None) is not None:
            self.config = Config(config_name=self.kwargs["config"])

        self.__tag_config = Undefined
        self.__test_thread_limit = Undefined
        self.__suite_thread_limit = Undefined
        self.__features = Undefined
        self.__components = Undefined
        self.__owners = Undefined
        self.__tests = Undefined
        self.__resources_mon = Undefined
        self.__html_report = Undefined
        self.__xml_report = Undefined
        self.__quiet = Undefined

        self.__print_settings()

    def __print_settings(self):

        LogJunkie.debug("============= Runtime Settings =============")
        LogJunkie.debug("Test Thread Limit: {value}:({type})".format(value=self.test_thread_limit,
                                                                    type=type(self.test_thread_limit)))
        LogJunkie.debug("Suite Thread Limit: {value}".format(value=self.suite_thread_limit))
        LogJunkie.debug("Features: {value}".format(value=self.features))
        LogJunkie.debug("Components: {value}".format(value=self.components))
        LogJunkie.debug("Owners: {value}".format(value=self.owners))
        LogJunkie.debug("Tests: {value}".format(value=self.tests))
        LogJunkie.debug("Tags: {value}".format(value=self.tags))
        LogJunkie.debug("Monitor Resources: {value}".format(value=self.monitor_resources))
        LogJunkie.debug("HTML Report: {value}:({type})".format(value=self.html_report, type=type(self.html_report)))
        LogJunkie.debug("XML Report: {value}:({type})".format(value=self.xml_report, type=type(self.xml_report)))
        LogJunkie.debug("Quiet: {value}:({type})".format(value=self.quiet, type=type(self.quiet)))
        LogJunkie.debug("============================================")

    def __get_value(self, key, default):
        """
        Generic method to resolve value to be used during runtime for a particular setting/property.
        1. Attempt to retrieve value from the explicitly passed in kwargs during Runner initiation
        2. If value is still __undefined__, attempt to retrieve value from the config __IF__ config was provided to
           the Runner during the initiation
        3. If value is still __undefined__, will use default values
        :param key: STRING, property key aka features, owners, test_multithreading_limit etc
        :param default: DATA VALUE, value to default to aka None, False, True, 1 etc
        :return: DATA VALUE
        """
        value = Undefined  # we start with __undefined__, because None is a valid value
        source = "DEFAULTS"

        # if we have kwargs, attempt to retrieve value for the key
        if self.kwargs is not None:
            value = self.kwargs.get(key, Undefined)
            source = "KWARGS"

        # if value is still __undefined__ and config provided, will check the config for a value to use
        if value is Undefined and self.config is not None:
            source = "DEFAULTS"
            if key in self.config.config.options("runtime"):
                value = self.config.get_value(key)
                if value is not Undefined:
                    value = ast.literal_eval(value)
                    source = "CONFIG @ {}".format(self.config.path)

        LogJunkie.debug("Setting: {setting} Source: {source}".format(setting=key, source=source))
        # if value is still __undefined__, will return default value
        return value if value is not Undefined else default

    @property
    def test_thread_limit(self):

        if self.__test_thread_limit is Undefined:
            self.__test_thread_limit = self.__get_value(key="test_multithreading_limit",
                                                        default=Settings.__DEFAULT_TEST_THREAD_LIMIT)
        return self.__test_thread_limit

    @property
    def suite_thread_limit(self):

        if self.__suite_thread_limit is Undefined:
            self.__suite_thread_limit = self.__get_value(key="suite_multithreading_limit",
                                                         default=Settings.__DEFAULT_SUITE_THREAD_LIMIT)
        return self.__suite_thread_limit

    @property
    def features(self):

        if self.__features is Undefined:
            self.__features = self.__get_value(key="features", default=Settings.__DEFAULT_FEATURES)
        return self.__features

    @property
    def components(self):

        if self.__components is Undefined:
            self.__components = self.__get_value(key="components", default=Settings.__DEFAULT_COMPONENTS)
        return self.__components

    @property
    def owners(self):

        if self.__owners is Undefined:
            self.__owners = self.__get_value(key="owners", default=Settings.__DEFAULT_OWNERS)
        return self.__owners

    @property
    def tests(self):
        if self.__tests is Undefined:
            self.__tests = self.__get_value(key="tests", default=Settings.__DEFAULT_TESTS)
        return self.__tests

    @property
    def tags(self):
        if self.__tag_config == Undefined:
            config = self.kwargs.get("tag_config", Undefined)
            if config is Undefined:
                config = {}
                properties = ["run_on_match_all", "run_on_match_any", "skip_on_match_all", "skip_on_match_any"]
                for prop in properties:
                    config.update({prop: self.__get_value(key=prop, default=Settings.__DEFAULT_TAGS)})
                self.__tag_config = config
            else:
                for prop, value in config.items():
                    if value is Undefined:
                        config.update({prop: self.__get_value(key=prop, default=Settings.__DEFAULT_TAGS)})
                self.__tag_config = config
        return self.__tag_config

    @property
    def monitor_resources(self):
        if self.__resources_mon is Undefined:
            self.__resources_mon = self.__get_value(key="monitor_resources", default=Settings.__DEFAULT_RESOURCE_MON)
        return self.__resources_mon

    @property
    def quiet(self):
        if self.__quiet is Undefined:
            self.__quiet = self.__get_value(key="quiet", default=Settings.__DEFAULT_QUIET)
        return self.__quiet

    @property
    def html_report(self):
        if self.__html_report is Undefined:
            self.__html_report = self.__get_value(key="html_report", default=Settings.__DEFAULT_HTML)
        if self.__html_report and not self.__html_report.endswith(".html"):
            raise BadParameters("\"html_report\" parameter requires full path with a file name and .html extension "
                                "for example: /var/www/html/my_report.html. For more info, see documentation: {link}"
                                .format(link=DocumentationLinks.HTML_REPORT))
        return self.__html_report

    @property
    def xml_report(self):
        if self.__xml_report is Undefined:
            self.__xml_report = self.__get_value(key="xml_report", default=Settings.__DEFAULT_XML)
        if self.__xml_report and not self.__xml_report.endswith(".xml"):
            raise BadParameters("\"xml_report\" parameter requires full path with a file name and .html extension "
                                "for example: /var/www/html/my_report.xml. For more info, see documentation: {link}"
                                .format(link=DocumentationLinks.XML_REPORT))
        return self.__xml_report



import copy
import inspect
import traceback
import sys
from test_junkie.decorators import DecoratorType
from test_junkie.constants import TestCategory
from test_junkie.errors import TestJunkieExecutionError, BadParameters
from test_junkie.metrics import ClassMetrics, TestMetrics, Aggregator


class _FuncEval:

    @staticmethod
    def eval_skip(obj):

        val = obj.get_skip()
        try:
            if not isinstance(val, bool):
                if inspect.isfunction(val):
                    val = val(meta=obj.get_meta()) if "meta" in inspect.getargspec(val).args else val()
                elif inspect.ismethod(val):
                    val = getattr(val.__self__, val.__name__)(meta=obj.get_meta()) \
                        if "meta" in inspect.getargspec(val).args \
                        else getattr(val.__self__, val.__name__)()
                else:
                    raise BadParameters("Unsupported data type used to pass parameters to the skip property in test: "
                                        "{}.{}".format(obj.get_function_module(), obj.get_function_name()))
                assert isinstance(val, bool), "You were using a function/unbound method to pass parameters to the skip"\
                                              " property in test {}.{} which must return a boolean value. "\
                                              "Got: {} instead of a boolean".format(obj.get_function_module(),
                                                                                    obj.get_function_name(), type(val))
        except Exception:
            traceback.print_exc()
            raise TestJunkieExecutionError("Encountered error while processing skip condition")
        return val

    @staticmethod
    def eval_params(params):

        try:
            if inspect.isfunction(params):
                return params()
            elif inspect.ismethod(params):
                return getattr(params.__self__, params.__name__)()
        except Exception:
            traceback.print_exc()
            raise TestJunkieExecutionError("Encountered error while processing parameters "
                                           "passed in via function or unbound method: {}".format(params))
        return params


class SuiteObject(object):

    def __init__(self, suite_definition):

        self.__suite_definition = suite_definition
        self.__listener = suite_definition["test_listener"](class_meta=suite_definition["class_meta"])
        self.__tests = []
        for test in suite_definition["suite_definition"].get(DecoratorType.TEST_CASE):
            self.__tests.append(TestObject(test))
        self.metrics = ClassMetrics()
        self.__rules = suite_definition["test_rules"](suite=copy.deepcopy(self))
        self.__instance = None

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    def get_suite_id(self):
        return self.get_kwargs().get("testjunkie_suite_id", 0)

    def get_decorated_definition(self, decorator_type):

        return self.__suite_definition["suite_definition"].get(decorator_type)

    def get_test_count(self):

        return len(self.__tests)

    def get_class_name(self):

        return self.__suite_definition["class_name"]

    def get_class_instance(self):

        if self.__instance is None:
            self.__instance = self.__suite_definition["class_object"]()
        return self.__instance

    def get_class_object(self):

        return self.__suite_definition["class_object"]

    def get_class_module(self):

        return self.get_class_object().__module__

    def _update_test_meta(self, parameter=None, suite_parameter=None, **kwargs):

        for frame in inspect.stack():
            try:
                getattr(self.get_class_object(), str(inspect.getframeinfo(frame[0]).function))
                test_function_name = str(inspect.getframeinfo(frame[0]).function)
                for test_object in self.get_test_objects():
                    if test_object.get_function_name() == test_function_name:
                        test_object.get_meta(parameter, suite_parameter).update(kwargs)
                        return True
                return False
            except:
                pass
        return False

    def _get_test_meta(self, parameter=None, suite_parameter=None):

        for frame in inspect.stack():
            try:
                getattr(self.get_class_object(), str(inspect.getframeinfo(frame[0]).function))
                test_function_name = str(inspect.getframeinfo(frame[0]).function)
                for test_object in self.get_test_objects():
                    if test_object.get_function_name() == test_function_name:
                        return test_object.get_meta(parameter, suite_parameter)
            except:
                pass

    def update_test_objects(self, tests):

        self.__tests = tests

    def get_test_objects(self):

        return self.__tests

    def get_skip(self):
        return self.__suite_definition.get("class_skip", False)

    def can_skip(self, features=None):

        can_skip = _FuncEval.eval_skip(self)
        if features is not None and can_skip is False:
            return not self.get_feature() in features
        return can_skip

    def get_retry_limit(self):

        return self.__suite_definition.get("class_retry", 1)

    def get_rules(self):

        return self.__rules

    def get_listener(self):

        return self.__listener

    def get_meta(self, copy_of_meta=False):

        if copy_of_meta:
            return copy.deepcopy(self.__suite_definition.get("class_meta", {}))
        else:
            return self.__suite_definition.get("class_meta", {})

    def get_unsuccessful_tests(self):

        unsuccessful_tests = []
        tests = self.get_test_objects()
        for test in tests:
            for class_param, metrics in test.metrics.get_metrics().items():
                for value in metrics.values():
                    if value["status"] in TestCategory.ALL_UN_SUCCESSFUL:
                        unsuccessful_tests.append(test)
        unsuccessful_tests = list(set(unsuccessful_tests))
        return unsuccessful_tests

    def has_unsuccessful_tests(self):

        tests = self.get_test_objects()
        for test in tests:
            for class_param, metrics in test.metrics.get_metrics().items():
                for value in metrics.values():
                    if value["status"] in TestCategory.ALL_UN_SUCCESSFUL:
                        return True
        return False

    def get_status(self):

        return self.metrics.get_metrics()["status"]

    def get_owner(self):

        return self.get_kwargs().get("owner", None)

    def get_parallel_restrictions(self):

        return self.get_kwargs().get("pr", [])

    def get_parameters(self, process_functions=False):
        if process_functions:
            parameters = self.__suite_definition["class_parameters"]
            eval_result = _FuncEval.eval_params(parameters)
            if eval_result is not None:
                self.__suite_definition["class_parameters"] = eval_result
        return self.__suite_definition["class_parameters"]

    def get_kwargs(self):

        return self.__suite_definition["decorator_kwargs"]

    def is_parallelized(self):

        return self.__suite_definition["parallelized"]

    def get_feature(self):

        return self.get_kwargs().get("feature", None)

    def get_priority(self):

        return self.get_kwargs().get("priority", None)

    def __get_average_metric(self, decorator, metric):

        from statistics import mean
        performance = self.metrics.get_metrics().get(decorator, {}).get(metric, None)
        return mean(performance) if performance else None

    def get_average_performance_of_after_class(self):
        return self.__get_average_metric(DecoratorType.AFTER_CLASS, "performance")

    def get_average_performance_of_before_class(self):
        return self.__get_average_metric(DecoratorType.BEFORE_CLASS, "performance")

    def get_average_performance_of_after_test(self):
        return self.__get_average_metric(DecoratorType.AFTER_TEST, "performance")

    def get_average_performance_of_before_test(self):
        return self.__get_average_metric(DecoratorType.BEFORE_TEST, "performance")

    def get_runtime(self):

        return self.metrics.get_metrics().get("runtime", None)

    def get_number_of_actual_retries(self):

        return self.metrics.get_metrics().get("retries", None)

    def get_data_by_tags(self):

        data = {"_totals_": Aggregator.get_template()}
        for test in self.get_test_objects():
            test_metrics = test.metrics.get_metrics()
            if test.get_tags():
                for tag in test.get_tags():
                    if tag not in data:
                        data.update({tag: Aggregator.get_template()})
                    Aggregator._update_report(data, test_metrics, tag)
            else:
                if None not in data:
                    data.update({None: Aggregator.get_template()})
                Aggregator._update_report(data, test_metrics, None)
        return data


class TestObject(object):

    def __init__(self, test_definition):

        self.__test_definition = test_definition

        self.metrics = TestMetrics()

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    def get_test_id(self):
        return self.get_kwargs().get("testjunkie_test_id", 0)

    def get_suite_id(self):
        return self.get_kwargs().get("testjunkie_suite_id", 0)

    def get_skip(self):
        return self.get_kwargs().get("skip", False)

    def can_skip(self):
        return _FuncEval.eval_skip(self)

    def skip_before_test_rule(self):
        return self.get_kwargs().get("skip_before_test_rule", False)

    def skip_before_test(self):
        return self.get_kwargs().get("skip_before_test", False)

    def skip_after_test_rule(self):
        return self.get_kwargs().get("skip_after_test_rule", False)

    def skip_after_test(self):
        return self.get_kwargs().get("skip_after_test", False)

    def get_owner(self):

        return self.get_kwargs().get("owner", None)

    def get_priority(self):

        return self.get_kwargs().get("priority", None)

    def get_parameters(self, process_functions=False):
        if process_functions:
            parameters = self.get_kwargs().get("parameters", [None])
            eval_result = _FuncEval.eval_params(parameters)
            if eval_result is not None:
                self.get_kwargs()["parameters"] = eval_result
        return self.get_kwargs().get("parameters", [None])

    def get_retry_limit(self):

        return self.get_kwargs().get("retry", 1)

    def get_function_name(self):

        return self.get_function_object().__name__

    def get_function_module(self):

        return self.get_function_object().__module__

    def get_function_object(self):

        return self.__test_definition["decorated_function"]

    def get_parallel_restrictions(self):

        return self.get_kwargs().get("pr", [])

    def get_tags(self):

        return self.get_kwargs().get("tags", [])

    def get_meta(self, parameter=None, class_parameter=None, copy_of_meta=False):

        string_param = str(parameter)
        string_class_param = str(class_parameter)

        if not self.get_kwargs().get("meta", {}):  # does not require meta to be defined in order to Meta.update it
            self.get_kwargs().update({"meta": {}})
        meta = self.get_kwargs()["meta"]

        if "original" not in meta:
            meta.update({"original": copy.deepcopy(meta)})

        if string_class_param not in meta:
            meta.update({string_class_param: {string_param: copy.deepcopy(meta["original"])}})

        if string_param not in meta[string_class_param]:
            meta[string_class_param].update({string_param: copy.deepcopy(meta["original"])})
        if copy_of_meta:
            return copy.deepcopy(meta[string_class_param][string_param])
        else:
            return meta[string_class_param][string_param]

    def get_kwargs(self):

        return self.__test_definition["decorator_kwargs"]

    def get_no_retry_on(self):

        return self.get_kwargs().get("no_retry_on", [])

    def get_retry_on(self):

        return self.get_kwargs().get("retry_on", [])

    def get_component(self):

        return self.get_kwargs().get("component", None)

    def is_parallelized(self):

        return self.get_kwargs().get("parallelized", True)

    def parallelized_parameters(self):

        return self.get_kwargs().get("parallelized_parameters", False)

    def accepts_test_and_suite_parameters(self):

        return "parameter" in inspect.getargspec(self.get_function_object()).args and \
               "suite_parameter" in inspect.getargspec(self.get_function_object()).args

    def accepts_test_parameters(self):

        return "parameter" in inspect.getargspec(self.get_function_object()).args

    def accepts_suite_parameters(self):

        return "suite_parameter" in inspect.getargspec(self.get_function_object()).args

    def __not_ran(self, param, class_param):
        """
        :param param: test parameter
        :param class_param: class parameter
        :return: BOOLEAN, True if test has not ran yet, False otherwise
        """
        return str(class_param) not in self.metrics.get_metrics() or \
            str(param) not in self.metrics.get_metrics()[str(class_param)]

    def is_qualified_for_retry(self, param=None, class_param=None):

        if self.__not_ran(param, class_param):
            return True
        test = self.metrics.get_metrics()[str(class_param)][str(param)]
        if test["status"] in TestCategory.ALL_UN_SUCCESSFUL:
            if self.get_no_retry_on() and type(test["exceptions"][-1]) in self.get_no_retry_on():
                return False
            elif self.get_retry_on() and type(test["exceptions"][-1]) not in self.get_retry_on():
                return False
            return True
        return False

    def get_status(self, param, class_param):

        if self.__not_ran(param, class_param):
            return None
        return self.metrics.get_metrics()[str(class_param)][str(param)]["status"]

    def get_number_of_actual_retries(self, param, class_param):

        if self.__not_ran(param, class_param):
            return 0
        return self.metrics.get_metrics()[str(class_param)][str(param)]["retry"]


class Limiter:

    # honor limits or not
    ACTIVE = True

    # truncation limits
    __DEFAULT_LIMIT = 3000
    EXCEPTION_MESSAGE_LIMIT = __DEFAULT_LIMIT
    TRACEBACK_LIMIT = __DEFAULT_LIMIT

    # throttling limits
    SUITE_THROTTLING = 0
    TEST_THROTTLING = 0

    @staticmethod
    def parse_exception_object(value):

        if Limiter.ACTIVE and value is not None:
            msg = value.message if sys.version_info[0] < 3 else str(value)
            if isinstance(msg, str) and len(msg) > Limiter.EXCEPTION_MESSAGE_LIMIT:
                value.message = "{} [. . .]".format(msg[:Limiter.EXCEPTION_MESSAGE_LIMIT])
        return value

    @staticmethod
    def parse_traceback(value):

        if Limiter.ACTIVE and value is not None:
            if isinstance(value, str) and len(value) > Limiter.TRACEBACK_LIMIT:
                value = "{} [. . .]".format(value[:Limiter.TRACEBACK_LIMIT])
        return value

    @staticmethod
    def get_suite_throttling():
        return Limiter.SUITE_THROTTLING if Limiter.ACTIVE else 0

    @staticmethod
    def get_test_throttling():
        return Limiter.TEST_THROTTLING if Limiter.ACTIVE else 0


class GroupRulesObject(object):

    def __init__(self, definition):

        self.definition = definition

    def run_after_group(self, suite):

        for group, definition in self.definition.items():
            if suite.get_class_object() in definition["suites"]:
                definition["suites"].remove(suite.get_class_object())
                if not definition["suites"] and DecoratorType.AFTER_GROUP in definition["rules"]:
                    for func in definition["rules"][DecoratorType.AFTER_GROUP]:
                        try:
                            func["decorated_function"]()
                        except Exception as error:
                            trace = traceback.format_exc()
                            return {"trace": trace, "exception": error}
        return None

    def run_before_group(self, suite, rule_type):

        for group, definition in self.definition.items():
            if suite.get_class_object() in definition["suites"] and rule_type in definition["rules"]:
                for func in list(definition["rules"][rule_type]):
                    try:
                        func["decorated_function"]()
                        definition["rules"].pop(rule_type)
                    except Exception as error:
                        trace = traceback.format_exc()
                        return {group: {"trace": trace, "exception": error, "definition": definition}}
        return None



import functools
import threading

from test_junkie.builder import Builder
from test_junkie.constants import DecoratorType


class Suite(object):

    def __init__(self, **decorator_kwargs):
        self.decorator_kwargs = decorator_kwargs

    def __call__(self, cls):

        Builder.build_suite_definitions(decorated_function=cls,
                                        decorator_kwargs=self.decorator_kwargs,
                                        decorator_type=DecoratorType.TEST_SUITE)
        return cls


class test(object):

    def __init__(self, **decorator_kwargs):
        self.decorator_kwargs = decorator_kwargs

    def __call__(self, decorated_function):

        Builder.build_suite_definitions(decorated_function=decorated_function,
                                        decorator_kwargs=self.decorator_kwargs,
                                        decorator_type=DecoratorType.TEST_CASE)
        return decorated_function


class beforeTest(object):

    def __init__(self, **decorator_kwargs):
        self.decorator_kwargs = decorator_kwargs

    def __call__(self, decorated_function):

        Builder.build_suite_definitions(decorated_function=decorated_function,
                                        decorator_kwargs=self.decorator_kwargs,
                                        decorator_type=DecoratorType.BEFORE_TEST)
        return decorated_function


class beforeClass(object):

    def __init__(self, **decorator_kwargs):
        self.decorator_kwargs = decorator_kwargs

    def __call__(self, decorated_function):

        Builder.build_suite_definitions(decorated_function=decorated_function,
                                        decorator_kwargs=self.decorator_kwargs,
                                        decorator_type=DecoratorType.BEFORE_CLASS)
        return decorated_function


class afterTest(object):

    def __init__(self, **decorator_kwargs):
        self.decorator_kwargs = decorator_kwargs

    def __call__(self, decorated_function):

        Builder.build_suite_definitions(decorated_function=decorated_function,
                                        decorator_kwargs=self.decorator_kwargs,
                                        decorator_type=DecoratorType.AFTER_TEST)
        return decorated_function


class afterClass(object):

    def __init__(self, **decorator_kwargs):
        self.decorator_kwargs = decorator_kwargs

    def __call__(self, decorated_function):

        Builder.build_suite_definitions(decorated_function=decorated_function,
                                        decorator_kwargs=self.decorator_kwargs,
                                        decorator_type=DecoratorType.AFTER_CLASS)
        return decorated_function


class GroupRules(object):

    def __init__(self, **decorator_kwargs):
        self.decorator_kwargs = decorator_kwargs

    def __call__(self, decorated_function):

        Builder.register_group_rules(decorated_function=decorated_function,
                                     decorator_kwargs=self.decorator_kwargs,
                                     decorator_type=DecoratorType.GROUP_RULES)
        return decorated_function


class afterGroup(object):

    def __init__(self, group, **decorator_kwargs):
        self.decorator_kwargs = decorator_kwargs
        self.group = group

    def __call__(self, decorated_function):

        Builder.add_group_rule(suites=self.group,
                               decorated_function=decorated_function,
                               decorator_kwargs=self.decorator_kwargs,
                               decorator_type=DecoratorType.AFTER_GROUP)
        return decorated_function


class beforeGroup(object):

    def __init__(self, group, **decorator_kwargs):
        self.decorator_kwargs = decorator_kwargs
        self.group = group

    def __call__(self, decorated_function):

        Builder.add_group_rule(suites=self.group,
                               decorated_function=decorated_function,
                               decorator_kwargs=self.decorator_kwargs,
                               decorator_type=DecoratorType.BEFORE_GROUP)
        return decorated_function


def synchronized(lock=threading.Lock()):
    def wrapper(f):
        @functools.wraps(f)
        def inner_wrapper(*args, **kw):
            with lock:
                return f(*args, **kw)
        return inner_wrapper
    return wrapper



from test_junkie.constants import TestCategory, DecoratorType
from test_junkie.runner import Runner
from tests.QualityManager import QualityManager
from tests.junkie_suites.BasicSuite import BasicSuite
from tests.junkie_suites.ExecutionSquence import ExecutionSequence1, ExecutionSequence2, ExecutionSequence3, \
    ExecutionSequence4
from tests.junkie_suites.ParametersSuite import ParametersSuite

runner = Runner([BasicSuite])
runner.run()
results = runner.get_executed_suites()
tests = results[0].get_test_objects()


def test_class_metrics():

    metrics = results[0].metrics.get_metrics()
    QualityManager.check_class_metrics(metrics,
                                       expected_status="fail",
                                       expected_beforeclass_exception_count=1,
                                       expected_beforeclass_exception_object=None,
                                       expected_beforeclass_performance_count=1,

                                       expected_afterclass_exception_count=1,
                                       expected_afterclass_exception_object=None,
                                       expected_afterclass_performance_count=1,

                                       expected_beforetest_exception_count=8,
                                       expected_beforetest_exception_object=None,
                                       expected_beforetest_performance_count=8,

                                       expected_aftertest_exception_count=8,
                                       expected_aftertest_exception_object=None,
                                       expected_aftertest_performance_count=8)


def test_failure():

    tested = False
    for test in tests:
        if test.get_function_name() == "failure":
            metrics = test.metrics.get_metrics()["None"]["None"]
            QualityManager.check_test_metrics(metrics,
                                              expected_status="fail",
                                              expected_exception=AssertionError)
            tested = True
    if not tested:
        raise Exception("Test did not run")


def test_error():

    tested = False
    for test in tests:
        if test.get_function_name() == "error":
            metrics = test.metrics.get_metrics()["None"]["None"]
            QualityManager.check_test_metrics(metrics,
                                              expected_status="error",
                                              expected_exception=Exception)
            tested = True
    if not tested:
        raise Exception("Test did not run")


def test_skip():

    tested = False
    for test in tests:
        if test.get_function_name() == "skip":
            metrics = test.metrics.get_metrics()["None"]["None"]
            QualityManager.check_test_metrics(metrics,
                                              expected_status="skip")
            tested = True
    if not tested:
        raise Exception("Test did not run")


def test_skip_function():

    tested = False
    for test in tests:
        if test.get_function_name() == "skip_function":
            metrics = test.metrics.get_metrics()["None"]["None"]
            QualityManager.check_test_metrics(metrics,
                                              expected_status="skip")
            tested = True
    if not tested:
        raise Exception("Test did not run")


def test_retry():

    tested = False
    for test in tests:
        if test.get_function_name() == "retry":
            metrics = test.metrics.get_metrics()["None"]["None"]
            QualityManager.check_test_metrics(metrics,
                                              expected_status="fail",
                                              expected_exception=AssertionError,
                                              expected_retry_count=2,
                                              expected_performance_count=2,
                                              expected_exception_count=2)
            tested = True
    if not tested:
        raise Exception("Test did not run")


def test_parameters():

    tested = False
    for test in tests:
        if test.get_function_name() == "parameters":
            properties = test.metrics.get_metrics()["None"]
            for param, metrics in properties.items():
                QualityManager.check_test_metrics(metrics,
                                                  expected_status="success",
                                                  expected_param=param)
            tested = True
    if not tested:
        raise Exception("Test did not run")


def test_parameters_plus_plus():
    runner = Runner([ParametersSuite])
    aggregator = runner.run()
    metrics = aggregator.get_basic_report()["tests"]
    assert metrics[TestCategory.SUCCESS] == 36
    assert metrics[TestCategory.IGNORE] == 4
    suites = runner.get_executed_suites()
    metrics = suites[0].metrics.get_metrics()
    QualityManager.check_class_metrics(metrics, expected_status="fail")


def test_execution_sequence1():
    runner = Runner([ExecutionSequence1])
    aggregator = runner.run()
    metrics = aggregator.get_basic_report()["tests"]
    assert metrics[TestCategory.SUCCESS] == 0
    assert metrics[TestCategory.FAIL] == 6
    suites = runner.get_executed_suites()
    metrics = suites[0].metrics.get_metrics()
    QualityManager.check_class_metrics(metrics, expected_status="fail",
                                       expected_retry_count=2,
                                       expected_beforetest_exception_count=20,
                                       expected_beforetest_performance_count=20,
                                       expected_beforetest_exception_object=AssertionError,
                                       expected_aftertest_exception_count=20,
                                       expected_aftertest_performance_count=20,
                                       expected_aftertest_exception_object=None
                                       )
    for test in suites[0].get_test_objects():
        test_metrics = test.metrics.get_metrics()
        for class_param, class_param_data in test_metrics.items():
            for param, param_data in class_param_data.items():
                expected_value = 4 if param_data["param"] is not None else 2
                assert len(param_data[DecoratorType.BEFORE_TEST]["exceptions"]) == expected_value
                assert len(param_data[DecoratorType.BEFORE_TEST]["tracebacks"]) == expected_value
                assert len(param_data[DecoratorType.BEFORE_TEST]["performance"]) == expected_value


def test_execution_sequence2():
    runner = Runner([ExecutionSequence2])
    aggregator = runner.run()
    metrics = aggregator.get_basic_report()["tests"]
    assert metrics[TestCategory.SUCCESS] == 0
    assert metrics[TestCategory.FAIL] == 6
    suites = runner.get_executed_suites()
    metrics = suites[0].metrics.get_metrics()
    QualityManager.check_class_metrics(metrics, expected_status="fail",
                                       expected_retry_count=2,
                                       expected_beforetest_exception_count=20,
                                       expected_beforetest_performance_count=20,
                                       expected_beforetest_exception_object=None,
                                       expected_aftertest_exception_count=20,
                                       expected_aftertest_performance_count=20,
                                       expected_aftertest_exception_object=AssertionError)
    for test in suites[0].get_test_objects():
        test_metrics = test.metrics.get_metrics()
        for class_param, class_param_data in test_metrics.items():
            for param, param_data in class_param_data.items():
                expected_value = 4 if param_data["param"] is not None else 2
                assert len(param_data[DecoratorType.AFTER_TEST]["exceptions"]) == expected_value
                assert len(param_data[DecoratorType.AFTER_TEST]["tracebacks"]) == expected_value
                assert len(param_data[DecoratorType.AFTER_TEST]["performance"]) == expected_value


def test_execution_sequence3():
    runner = Runner([ExecutionSequence3])
    aggregator = runner.run()
    metrics = aggregator.get_basic_report()["tests"]
    assert metrics[TestCategory.SUCCESS] == 5
    assert metrics[TestCategory.FAIL] == 1
    suites = runner.get_executed_suites()
    metrics = suites[0].metrics.get_metrics()
    QualityManager.check_class_metrics(metrics, expected_status="fail",
                                       expected_retry_count=2,
                                       expected_beforetest_exception_count=7,
                                       expected_beforetest_performance_count=7,
                                       expected_beforetest_exception_object=None,
                                       expected_aftertest_exception_count=7,
                                       expected_aftertest_performance_count=7,
                                       expected_aftertest_exception_object=None)
    for test in suites[0].get_test_objects():
        test_metrics = test.metrics.get_metrics()
        for class_param, class_param_data in test_metrics.items():
            for param, param_data in class_param_data.items():
                status = test.get_status(class_param, param)
                expected_value = 2 if status == TestCategory.FAIL else 1
                assert len(param_data[DecoratorType.AFTER_TEST]["exceptions"]) == expected_value
                assert len(param_data[DecoratorType.AFTER_TEST]["tracebacks"]) == expected_value
                assert len(param_data[DecoratorType.AFTER_TEST]["performance"]) == expected_value
                assert len(param_data[DecoratorType.BEFORE_TEST]["exceptions"]) == expected_value
                assert len(param_data[DecoratorType.BEFORE_TEST]["tracebacks"]) == expected_value
                assert len(param_data[DecoratorType.BEFORE_TEST]["performance"]) == expected_value


def test_execution_sequence4():
    runner = Runner([ExecutionSequence4])
    aggregator = runner.run()
    metrics = aggregator.get_basic_report()["tests"]
    assert metrics[TestCategory.SUCCESS] == 6
    assert metrics[TestCategory.FAIL] == 0
    suites = runner.get_executed_suites()
    metrics = suites[0].metrics.get_metrics()
    QualityManager.check_class_metrics(metrics, expected_status="success")
    for test in suites[0].get_test_objects():
        test_metrics = test.metrics.get_metrics()
        for class_param, class_param_data in test_metrics.items():
            for param, param_data in class_param_data.items():
                expected_value = 0
                assert len(param_data[DecoratorType.AFTER_TEST]["exceptions"]) == expected_value
                assert len(param_data[DecoratorType.AFTER_TEST]["tracebacks"]) == expected_value
                assert len(param_data[DecoratorType.AFTER_TEST]["performance"]) == expected_value
                assert len(param_data[DecoratorType.BEFORE_TEST]["exceptions"]) == expected_value
                assert len(param_data[DecoratorType.BEFORE_TEST]["tracebacks"]) == expected_value
                assert len(param_data[DecoratorType.BEFORE_TEST]["performance"]) == expected_value



# TODO overhaul errors and add documentation links


class TestJunkieExecutionError(Exception):

    def __init__(self, message):
        Exception.__init__(self, message)


class TestListenerError(TestJunkieExecutionError):

    def __init__(self, message):
        TestJunkieExecutionError.__init__(self, message)


class ConfigError(TestJunkieExecutionError):

    def __init__(self, message):
        TestJunkieExecutionError.__init__(self, message)


class BadParameters(TestJunkieExecutionError):

    def __init__(self, message):
        TestJunkieExecutionError.__init__(self, message)


class BadCliParameters(TestJunkieExecutionError):

    def __init__(self, message):
        TestJunkieExecutionError.__init__(self, message)


class BadSignature(Exception):

    def __init__(self, message):
        Exception.__init__(self, message)



from test_junkie.decorators import test, beforeClass, Suite
from tests.junkie_suites.TestListener import TestListener


@Suite(retry=2, listener=TestListener)
class BeforeClassAssertionSuite:

    @beforeClass()
    def before_class(self):
        raise AssertionError("Assertion Error in before class")

    @test(retry=2)
    def ignore_1(self):
        pass

    @test(retry=2)
    def ignore_2(self):
        pass



from difflib import SequenceMatcher
from test_junkie.constants import DocumentationLinks


class Analyzer:

    def __init__(self, monitoring_enabled=False, multi_threading_enabled=False):

        self.multi_threading_enabled = multi_threading_enabled
        self.__analysis = {"missing_definitions": {},
                           "traceback_insights": {},
                           "resources": {"monitoring_enabled": monitoring_enabled,
                                         "cpu": {"high": [], "medium": []},
                                         "mem": {"high": [], "medium": []}},
                           "time_lost_retrying": [],
                           "stable": True}

    def update_resources(self, cpu, mem):

        for level, value in {"high": 98, "medium": 70}.items():
            if cpu >= value:
                self.__analysis["resources"]["cpu"][level].append(cpu)
            if mem >= value:
                self.__analysis["resources"]["mem"][level].append(mem)

    def analyze(self, test_id, tracebacks, performance):
        """
        This function analyzes data for a specific test case against other known data from different test cases
        :param test_id: INT id of the test
        :param tracebacks: LIST of strings of the tracebacks
        :param performance: LIST of floats
        :return: None
        """
        index = 0
        for traceback in tracebacks:

            if index > 0:
                self.__analysis["time_lost_retrying"].append(performance[index])

            if traceback and traceback not in self.__analysis["traceback_insights"]:
                self.__analysis["stable"] = False
                self.__analysis["traceback_insights"].update({traceback: {"similar": [],
                                                                          "exact": [],
                                                                          "test_id": test_id}})
                for key in self.__analysis["traceback_insights"].keys():
                    if key != traceback:
                        if Analyzer.is_similar(traceback, key):
                            self.__analysis["traceback_insights"][key]["similar"].append(test_id)
            elif traceback and index == 0:
                self.__analysis["traceback_insights"][traceback]["exact"].append(test_id)

            index += 1

    @property
    def analysis(self):
        """
        This function does a final analysis across the dataset at the time and returns usable HTML for the report
        :return: LIST of STRINGs with HTML ready to go
        """
        from test_junkie.reporter.html_reporter import Reporter

        analysis = []
        # analyzing retries
        retry_link = "<a ref='noopener' target='_blank' href='{}'>retries <i class='fas fa-external-link-alt'></i></a>"\
                     .format(DocumentationLinks.RETRY)
        if self.__analysis["time_lost_retrying"]:
            analysis.append("Total of {retries} {link} cost you {seconds} seconds."
                            .format(retries=len(self.__analysis["time_lost_retrying"]),
                                    seconds=Reporter.total_up(self.__analysis["time_lost_retrying"]),
                                    link=retry_link))
        else:
            if self.__analysis["stable"]:
                analysis.append("All of your tests are stable and thus no time was lost on {link}."
                                .format(link=retry_link))

        # analyzing tracebacks
        if self.__analysis["traceback_insights"]:
            if not self.__analysis["stable"]:
                analysis.append("There are {tracebacks} unique tracebacks across all of the unsuccessful tests."
                                .format(tracebacks=len(self.__analysis["traceback_insights"])))
                ones_to_report_on = {}
                for traceback, category in self.__analysis["traceback_insights"].items():
                    # tb = traceback.replace("\n", "<br>").replace("    ", "&emsp;")
                    test_id = category["test_id"]
                    if test_id not in ones_to_report_on:
                        ones_to_report_on.update({test_id: {"data": category, "traceback": traceback}})
                    elif len(category["similar"]) > len(ones_to_report_on[test_id]["data"]["similar"]):
                        ones_to_report_on.update({test_id: {"data": category, "traceback": traceback}})

                if ones_to_report_on:
                    for test_id, data in ones_to_report_on.items():
                        if len(data["data"]["similar"]) > 0:
                            tb = Reporter.escape(data["traceback"], quote=True)

                            analysis.append(
                                "{similar_count} test failures due to a similar "
                                "<span data-html='true' data-tooltip='<xmp>{traceback}</xmp>' "
                                "class='tooltipped traceback-tooltip'>"
                                "&nbsp;traceback <i class='fas fa-project-diagram'></i>&nbsp;</span>"
                                .format(traceback=tb, similar_count=len(data["data"]["similar"]) + 1))

        # analyzing CPU/MEM
        if self.__analysis["resources"]["monitoring_enabled"]:
            link = "See documentation for <a ref='noopener' target='_blank' href='{}'>Parallel Execution " \
                   "<i class='fas fa-external-link-alt'></i></a>".format(DocumentationLinks.THREADING)
            if len(self.__analysis["resources"]["cpu"]["high"]) > 10:
                msg = "CPU has spiked {} times to critical levels over the course of test execution which may have "\
                      "impacted some of the tests. You may want to shut down any unnecessary background processes{}"\
                      .format(len(self.__analysis["resources"]["cpu"]["high"]),
                              "and/or reduce thread allocation. {}".format(link)
                              if self.multi_threading_enabled else ".")
                analysis.append(msg)
            else:
                if not self.multi_threading_enabled:
                    msg = "Consider enabling multi-threading."
                else:
                    msg = "More of the resources can be utilized, consider allocating more " \
                          "threads. {}".format(link)
                analysis.append(msg)

        return analysis

    @staticmethod
    def is_similar(string_a, string_b):
        if SequenceMatcher(None, string_a, string_b).ratio() >= 0.70:
            return True
        return False

    def is_missing_definitions(self, data):

        pass





class Listener(object):

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    @staticmethod
    def __process_event(**kwargs):
        if kwargs.get("custom_function", None) is not None:
            if kwargs.get("error", None) is None:
                kwargs.get("custom_function")(properties=kwargs.get("properties"))
            else:
                kwargs.get("custom_function")(properties=kwargs.get("properties", None),
                                              exception=kwargs.get("error", None),
                                              trace=kwargs.get("trace", None))

    def on_cancel(self, **kwargs):
        Listener.__process_event(**kwargs)

    def on_success(self, **kwargs):
        Listener.__process_event(**kwargs)

    def on_failure(self, **kwargs):
        Listener.__process_event(**kwargs)

    def on_skip(self, **kwargs):
        Listener.__process_event(**kwargs)

    def on_error(self, **kwargs):
        Listener.__process_event(**kwargs)

    def on_ignore(self, **kwargs):
        Listener.__process_event(**kwargs)

    def on_in_progress(self, **kwargs):
        Listener.__process_event(**kwargs)

    def on_complete(self, **kwargs):
        Listener.__process_event(**kwargs)

    def on_before_class_error(self, **kwargs):
        Listener.__process_event(**kwargs)

    def on_before_class_failure(self, **kwargs):
        Listener.__process_event(**kwargs)

    def on_after_class_error(self, **kwargs):
        Listener.__process_event(**kwargs)

    def on_after_class_failure(self, **kwargs):
        Listener.__process_event(**kwargs)

    def on_class_skip(self, **kwargs):
        Listener.__process_event(**kwargs)

    def on_class_cancel(self, **kwargs):
        Listener.__process_event(**kwargs)

    def on_class_in_progress(self, **kwargs):
        Listener.__process_event(**kwargs)

    def on_class_complete(self, **kwargs):
        Listener.__process_event(**kwargs)

    def on_class_ignore(self, **kwargs):
        Listener.__process_event(**kwargs)

    def on_before_group_failure(self, **kwargs):
        Listener.__process_event(**kwargs)

    def on_before_group_error(self, **kwargs):
        Listener.__process_event(**kwargs)

    def on_after_group_failure(self, **kwargs):
        Listener.__process_event(**kwargs)

    def on_after_group_error(self, **kwargs):
        Listener.__process_event(**kwargs)



import threading
from test_junkie.decorators import synchronized


def meta(**kwargs):
    return kwargs


class Meta:

    @staticmethod
    @synchronized(threading.Lock())
    def update(suite, parameter=None, suite_parameter=None, **kwargs):
        from test_junkie.builder import Builder
        suite_objects = Builder.get_execution_roster().values()
        for suite_object in suite_objects:
            if suite.__class__ == suite_object.get_class_object():
                if suite_object._update_test_meta(parameter, suite_parameter, **kwargs) is True:
                    return

    @staticmethod
    @synchronized(threading.Lock())
    def get_meta(suite, parameter=None, suite_parameter=None):
        from test_junkie.builder import Builder
        suite_objects = Builder.get_execution_roster().values()
        for suite_object in suite_objects:
            if suite.__class__ == suite_object.get_class_object():
                m = suite_object._get_test_meta(parameter, suite_parameter)
                if m is not None:
                    return m



import argparse
import ast
from appdirs import *

from test_junkie.constants import CliConstants, Undefined
from test_junkie.cli.cli import CliUtils


class Config:

    def __init__(self, config_name):

        if os.path.isfile(config_name):
            self.path = config_name
        else:
            self.path = "{root}{sep}{file}".format(root=Config.get_root_dir(), file=config_name, sep=os.sep)

        if not os.path.exists(Config.get_root_dir()):
            os.makedirs(Config.get_root_dir())
        if not os.path.exists(self.path):
            self.restore()
        self.config = self.__get_parser()

    def remove(self):
        """
        Will remove config from file storage
        :return: None
        """
        if os.path.exists(self.path):
            os.remove(self.path)

    def restore(self, data=None):
        """
        Overrides the config with provided data or defaults
        :param data: STRING, Optional data to put in the config
        :return: None
        """
        self.remove()
        with open(self.path, "w+") as doc:
            doc.write(data if data else CliConstants.DEFAULTS)

    def set_value(self, option, value):

        self.config.set('runtime', option, str(value))
        with open(self.path, 'w+') as doc:
            self.config.write(doc)

    def get_value(self, option, default=Undefined):

        if sys.version_info[0] < 3:
            # Python 2
            value = self.config.get("runtime", option, default)
        else:
            # Python 3, module is not backwards compatible and fallback has to be explicitly assigned
            value = self.config.get("runtime", option, fallback=default)
        return value

    def read(self):

        with open(self.path, 'r') as doc:
            return doc.read()

    @staticmethod
    def get_root_dir():
        """
        :return: STRING, root directory for TJ to store its configs and other assets
        """
        return user_data_dir("Test-Junkie")

    @staticmethod
    def get_config_path(config_name):
        """
        :param config_name: STRING, name of the config that you want to get the path for
        :return: STRING, path to the requested config
        """
        return "{root}{sep}{name}".format(root=Config.get_root_dir(), name=config_name, sep=os.sep)

    def __get_parser(self):
        """
        :param path: STRING, path to the config file
        :return: ConfigParser object
        """
        if sys.version_info[0] < 3:
            # Python 2
            import ConfigParser as configparser
        else:
            # Python 3, module was renamed to configparser
            import configparser
        config = configparser.ConfigParser()
        config.read(self.path)
        return config


class CliConfig:
    """
    only use for cli, do not use for parsing the config when running tests
    """

    def __init__(self, config_name, command, args):

        self.args = args
        self.config = Config(config_name=config_name)
        getattr(self, command)()

    def __restore_value(self, option, value):

        try:
            self.config.set_value(option, value)
            print("[{status}]\t{option}={value}".format(
                status=CliUtils.format_color_string(value="OK", color="green"), option=option, value=value))
        except:
            print("[{status}]\tUnexpected error occurred during update of {setting}={value}"
                  .format(status=CliUtils.format_color_string(value="ERROR", color="red"), setting=option, value=value))
            CliUtils.print_color_traceback()
            exit(120)

    def __print_value(self, option):

        if option in self.config.config.options("runtime"):
            value = self.config.get_value(option)
            if value != Undefined:
                print("{option}={value}".format(option=option, value=ast.literal_eval(value)))

    def update(self):

        parser = argparse.ArgumentParser(description="Update configuration settings for individual properties",
                                         usage="tj config update [OPTIONS]")
        CliUtils.add_standard_tj_args(parser)
        args = parser.parse_args(self.args[3:])
        if not self.args[3:]:
            print("[{status}]\tWhat do you want to update?\n".format(
                status=CliUtils.format_color_string(value="ERROR", color="red")))
            parser.print_help()
            return
        for option, value in args.__dict__.items():
            if value is not Undefined:
                self.__restore_value(option, value)
        print("[{status}]\tRestore original value with tj config restore [OPTION]. tj config restore -h for more info."
              .format(status=CliUtils.format_color_string(value="TIP", color="blue")))

    def show(self):

        parser = argparse.ArgumentParser(description='Display current configuration for Test-Junkie',
                                         usage="tj config show [OPTIONS]")
        parser.add_argument("-a", "--all", action="store_true", default=False,
                            help="Will restore all config settings to its default values")
        CliUtils.add_standard_boolean_tj_args(parser)
        args = parser.parse_args(self.args[3:])
        if args.all or not self.args[3:]:
            print("Config is located at: {path}\n"
                  .format(path=CliUtils.format_color_string(value=self.config.path, color="green")))
            with open(self.config.path, "r") as cfg:
                print(cfg.read())
            return
        for option, value in args.__dict__.items():
            if value is True:
                self.__print_value(option)

    def restore(self):

        parser = argparse.ArgumentParser(description='Restore config settings to it\'s original values',
                                         usage="tj config restore [OPTIONS]")
        parser.add_argument("-a", "--all", action="store_true", default=False,
                            help="Will restore all config settings to its default values")
        CliUtils.add_standard_boolean_tj_args(parser)
        args = parser.parse_args(self.args[3:])
        if args.all:
            self.config.restore()
            print("[{status}] Config restored to default settings!"
                  .format(status=CliUtils.format_color_string(value="OK", color="green")))
            return
        if not self.args[3:]:
            print("[{status}]\tWhat do you want to restore?\n".format(
                status=CliUtils.format_color_string(value="ERROR", color="red")))
            parser.print_help()
            return
        for option, value in args.__dict__.items():
            if value is True:
                self.__restore_value(option, None)



import io
import logging
import sys
from contextlib import contextmanager

from test_junkie.decorators import synchronized


class LogJunkie:

    __LOGGER = None
    __ENABLED = False
    __PRIORITY = logging.ERROR
    __NAME = "TestJunkieLogger"
    __FORMAT = "%(asctime)s [%(levelname)s] (%(filename)s, %(funcName)s(), %(lineno)d - %(threadName)s) :: %(message)s"

    @staticmethod
    def enable_logging(level):

        LogJunkie.__ENABLED = True
        LogJunkie.__PRIORITY = level

    @staticmethod
    def disable_logging():

        LogJunkie.__ENABLED = False

    @staticmethod
    @synchronized()
    def __get_logger():

        if LogJunkie.__LOGGER is None:
            stderr_handler = logging.StreamHandler()
            stderr_handler.setFormatter(logging.Formatter(LogJunkie.__FORMAT))
            LogJunkie.__LOGGER = logging.getLogger(LogJunkie.__NAME)
            LogJunkie.__LOGGER.addHandler(stderr_handler)
            LogJunkie.__LOGGER.setLevel(LogJunkie.__PRIORITY)
        return LogJunkie.__LOGGER

    @staticmethod
    def info(msg):

        if LogJunkie.__ENABLED:
            LogJunkie.__get_logger().info(msg)

    @staticmethod
    def debug(msg):

        if LogJunkie.__ENABLED:
            LogJunkie.__get_logger().debug(msg)

    @staticmethod
    def error(msg, exc_info=False):

        if LogJunkie.__ENABLED:
            LogJunkie.__get_logger().error(msg, exc_info=exc_info)

    @staticmethod
    def warn(msg):

        if LogJunkie.__ENABLED:
            LogJunkie.__get_logger().warning(msg)


@contextmanager
def suppressed_stdout(suppress=False):
    if suppress:
        original_stdout = sys.stdout
        if sys.version_info[0] < 3:
            sys.stdout = io.BytesIO()  # works with python 2
        else:
            sys.stdout = io.StringIO()  # works with python 3
        original_level = logging.root.manager.disable
        logging.disable(logging.ERROR)
        try:
            yield
        finally:
            sys.stdout = original_stdout
            logging.disable(original_level)
    else:
        yield



from test_junkie.debugger import LogJunkie
from test_junkie.errors import BadParameters
from test_junkie.runner import Runner

LogJunkie.enable_logging(10)

LogJunkie.debug("1")
LogJunkie.info("2")
LogJunkie.warn("3")
LogJunkie.error("4")
LogJunkie.disable_logging()


def test_bad_runner_initiation1():
    try:
        Runner(suites=None)
        raise AssertionError("Must have raised exception because bad args were passed in")
    except Exception as error:
        assert isinstance(error, BadParameters), "Type of exception is incorrect"



from test_junkie.runner import Runner
from tests.QualityManager import QualityManager
from tests.junkie_suites.BeforeClassExceptionSuite import BeforeClassExceptionSuite

runner = Runner([BeforeClassExceptionSuite])
runner.run()
results = runner.get_executed_suites()


def test_class_metrics():

    metrics = results[0].metrics.get_metrics()
    QualityManager.check_class_metrics(metrics,
                                       expected_status="fail",
                                       expected_retry_count=2,
                                       expected_beforeclass_exception_count=2,
                                       expected_beforeclass_exception_object=Exception,
                                       expected_beforeclass_performance_count=2)


def test_test_metrics():

    assert results[0].get_test_objects()
    for test in results[0].get_test_objects():

        metrics = test.metrics.get_metrics()["None"]["None"]
        QualityManager.check_test_metrics(metrics,
                                          expected_status="ignore",
                                          expected_exception_count=2,
                                          expected_performance_count=2,
                                          expected_retry_count=2,
                                          expected_exception=Exception)



from test_junkie.decorators import Suite, beforeClass, afterClass, test
from tests.junkie_suites.TestListener import TestListener


@Suite(retry=2, listener=TestListener)
class AfterClassAssertionSuite:

    @beforeClass()
    def before_class(self):
        pass

    @afterClass()
    def after_class(self):
        raise AssertionError("Assertion Error in after class")

    @test(retry=2)
    def pass_1(self):
        pass

    @test(retry=2)
    def pass_2(self):
        pass





class TestCategory:

    SUCCESS = "success"
    FAIL = "fail"
    IGNORE = "ignore"
    ERROR = "error"
    SKIP = "skip"
    CANCEL = "cancel"
    ALL = [SUCCESS, FAIL, IGNORE, ERROR, SKIP, CANCEL]
    ALL_UN_SUCCESSFUL = [FAIL, IGNORE, ERROR]


class SuiteCategory:

    SUCCESS = "success"
    FAIL = "fail"
    SKIP = "skip"
    CANCEL = "cancel"
    IGNORE = "ignore"
    ALL = [SUCCESS, FAIL, SKIP, CANCEL, IGNORE]
    ALL_UN_SUCCESSFUL = [FAIL, IGNORE]

    """
    This status means something failed outside of Test Junkie,
    like custom event listener function was missing required arguments in the signature.
    This status is used as fallback and is not set by Test Junkie during test execution.
    Tests that error out, suite should be marked with the FAIL status.
    Thus, this status is not part of the ALL nor ALL_UN_SUCCESSFUL list and should stay that way
    """
    ERROR = "error"


class DecoratorType:

    TEST_SUITE = "testSuite"
    BEFORE_CLASS = "beforeClass"
    BEFORE_TEST = "beforeTest"
    AFTER_TEST = "afterTest"
    AFTER_CLASS = "afterClass"
    TEST_CASE = "testCase"
    GROUP_RULES = "groupRules"
    BEFORE_GROUP = "beforeGroup"
    AFTER_GROUP = "afterGroup"


class Event:

    ON_SUCCESS = 1
    ON_FAILURE = 2
    ON_ERROR = 3
    ON_SKIP = 4
    ON_IGNORE = 5
    ON_CANCEL = 6
    ON_CLASS_SKIP = 7
    ON_CLASS_CANCEL = 8
    ON_BEFORE_CLASS_ERROR = 9
    ON_BEFORE_CLASS_FAIL = 10
    ON_AFTER_CLASS_ERROR = 11
    ON_AFTER_CLASS_FAIL = 12
    ON_CLASS_IN_PROGRESS = 13
    ON_CLASS_COMPLETE = 14
    ON_CLASS_IGNORE = 15
    ON_BEFORE_GROUP_FAIL = 16
    ON_BEFORE_GROUP_ERROR = 17
    ON_AFTER_GROUP_FAIL = 18
    ON_AFTER_GROUP_ERROR = 19
    ON_IN_PROGRESS = 20
    ON_COMPLETE = 21


class DocumentationLinks:

    DOMAIN = "https://www.test-junkie.com"
    SUITE_DECORATOR = "{}/documentation/#suite".format(DOMAIN)
    TEST_DECORATOR = "{}/documentation/#test".format(DOMAIN)
    ON_CLASS_IGNORE = "{}/documentation/#on_class_ignore".format(DOMAIN)
    ON_TEST_IGNORE = "{}/documentation/#on_ignore".format(DOMAIN)
    PARAMETERIZED_TESTS = "{}/documentation/#parameters".format(DOMAIN)
    LISTENERS = "{}/documentation/#listeners".format(DOMAIN)
    TAGS = "{}/documentation/#tags".format(DOMAIN)
    RUNNER_OBJECT = "{}/documentation/#runner".format(DOMAIN)
    GROUP_RULES = "{}/documentation/#GroupRules".format(DOMAIN)
    THREADING = "{}/documentation/#parallel_execution".format(DOMAIN)
    HTML_REPORT = "{}/documentation/#html_report".format(DOMAIN)
    XML_REPORT = "{}/documentation/#xml_report".format(DOMAIN)
    RETRY = "{}/documentation/#retry".format(DOMAIN)
    FEATURES = "{}/documentation/#features".format(DOMAIN)
    COMPONENTS = "{}/documentation/#components".format(DOMAIN)
    ASSIGNEES = "{}/documentation/#assignees".format(DOMAIN)

    COVERAGE_CONFIG_FILE = "https://coverage.readthedocs.io/en/v4.5.x/config.html"


class Color:

    SUCCESS = "#12d479"
    FAIL = "#fcd75f"
    ERROR = "#ff7651"
    IGNORE = "#cce4eb"
    SKIP = "#34bff5"
    CANCEL = "#f19def"

    MAPPING = {TestCategory.SUCCESS: SUCCESS,
               TestCategory.FAIL: FAIL,
               TestCategory.ERROR: ERROR,
               TestCategory.IGNORE: IGNORE,
               TestCategory.SKIP: SKIP,
               TestCategory.CANCEL: CANCEL}


class CliConstants:

    TJ_CONFIG_NAME = ".tj.cfg"

    DEFAULTS = """
[runtime]
test_multithreading_limit=None
suite_multithreading_limit=None
html_report=None
xml_report=None
monitor_resources=None
tests=None
features=None
components=None
owners=None
run_on_match_all=None
run_on_match_any=None
skip_on_match_all=None
skip_on_match_any=None
sources=None
quiet=None
code_cov=None
cov_rcfile=None
guess_root=None
"""


class Undefined(object):
    def __init__(self):
        pass



import argparse
import sys
import traceback
import pkg_resources

from test_junkie.cli.cli_audit import CliAudit
from test_junkie.constants import DocumentationLinks, CliConstants, Undefined
from test_junkie.errors import BadCliParameters
from colorama import Fore, Style


class Cli(object):

    def __init__(self):
        parser = argparse.ArgumentParser(
            description="",
            usage="""tj COMMAND

Modern Testing Framework

Commands:
run\t Run tests in any directory (recursive) 
audit\t Audit your tests in any directory (recursive) 
config\t Configure Test Junkie
version\t Display current version

Use: tj COMMAND -h to display COMMAND specific help
""")
        parser.add_argument('command', help='command to run')
        args = parser.parse_args(sys.argv[1:2])
        if not hasattr(self, args.command):
            print("[{status}]\t\'{command}\' is not a test-junkie command\n".format(
                status=CliUtils.format_color_string(value="ERROR", color="red"), command=args.command))
            parser.print_help()
            exit(1)
        if args.command:
            getattr(self, args.command)()

    def run(self):
        parser = argparse.ArgumentParser(description='Run tests from command line',
                                         usage="tj run [OPTIONS]")

        parser.add_argument("-x", "--suites", nargs="+", default=None,
                            help="Test Junkie will only run suites provided, "
                                 "given that they are found in the SOURCE")

        parser.add_argument("-v", "--verbose", action="store_true", default=False,
                            help="Enables Test Junkie's logs for debugging purposes")

        CliUtils.add_standard_tj_args(parser)

        args = parser.parse_args(sys.argv[2:])

        if args.verbose:
            from test_junkie.debugger import LogJunkie
            LogJunkie.enable_logging(10)

        from test_junkie.cli.cli_runner import CliRunner
        try:
            tj = CliRunner(sources=args.sources, ignore=[".git"], suites=args.suites,
                           code_cov=args.code_cov, cov_rcfile=args.cov_rcfile, guess_root=args.guess_root)
            tj.scan()
        except BadCliParameters as error:
            print("[{status}] {error}".format(status=CliUtils.format_color_string("ERROR", "red"), error=error))
            return
        tj.run_suites(args)

    def audit(self):
        parser = argparse.ArgumentParser(description='Scan and display aggregated and/or filtered test information',
                                         usage="""tj audit [COMMAND] [OPTIONS]

Aggregate, pivot, and display data about your tests.

Commands:
suites\t\t Pivot test information from suite's perspective
features\t Pivot test information from feature's perspective
components\t Pivot test information from component's perspective
tags\t\t Pivot test information from tag's perspective
owners\t\t Pivot test information from owner's perspective

usage: tj audit [COMMAND] [OPTIONS]
""")
        parser.add_argument('command', help='command to run')

        parser.add_argument("--by-components", action="store_true", default=False,
                            help="Present aggregated data broken down by components")

        parser.add_argument("--by-features", action="store_true", default=False,
                            help="Present aggregated data broken down by features")

        parser.add_argument("--no-rules", action="store_true", default=False,
                            help="Aggregate data only for suites that do not have any rules set")

        parser.add_argument("--no-listeners", action="store_true", default=False,
                            help="Aggregate data only for suites that do not have any event listeners set")

        parser.add_argument("--no-suite-retries", action="store_true", default=False,
                            help="Aggregate data only for suites that do not have retries set")

        parser.add_argument("--no-test-retries", action="store_true", default=False,
                            help="Aggregate data only for tests that do not have retries set")

        parser.add_argument("--no-suite-meta", action="store_true", default=False,
                            help="Aggregate data only for suites that do not have any meta information set")

        parser.add_argument("--no-test-meta", action="store_true", default=False,
                            help="Aggregate data only for tests that do not have any meta information set")

        parser.add_argument("--no-owners", action="store_true", default=False,
                            help="Aggregate data only for tests that do not have any owners defined")

        parser.add_argument("--no-features", action="store_true", default=False,
                            help="Aggregate data only for suites that do not have features defined")

        parser.add_argument("--no-components", action="store_true", default=False,
                            help="Aggregate data only for tests that do not have any components defined")

        parser.add_argument("--no-tags", action="store_true", default=False,
                            help="Aggregate data only for tests that do not have tags defined")

        parser.add_argument("-x", "--suites", nargs="+", default=None,
                            help="Test Junkie will only run suites provided, "
                                 "given that they are found in the SOURCE")

        parser.add_argument("-v", "--verbose", action="store_true", default=False,
                            help="Enables Test Junkie's logs for debugging purposes")

        CliUtils.add_standard_tj_args(parser, audit=True)

        if len(sys.argv) >= 3:
            args = parser.parse_args(sys.argv[2:])
            command = args.command
            if command not in ["suites", "features", "components", "tags", "owners"]:
                print("[{status}]\t\'{command}\' is not a test-junkie command\n".format(
                    status=CliUtils.format_color_string(value="ERROR", color="red"),
                    command=command))
                parser.print_help()
                exit(120)
            else:
                if args.verbose:
                    from test_junkie.debugger import LogJunkie
                    LogJunkie.enable_logging(10)

                from test_junkie.cli.cli_runner import CliRunner
                try:
                    tj = CliRunner(sources=args.sources, ignore=[".git"], suites=args.suites,
                                   guess_root=args.guess_root)
                    tj.scan()
                except BadCliParameters as error:
                    print("[{status}] {error}".format(status=CliUtils.format_color_string("ERROR", "red"), error=error))
                    return
                aggregator = CliAudit(suites=tj.suites, args=args)
                aggregator.aggregate()
                aggregator.print_results()
                return
        else:
            print("[{status}]\tDude, what do you want to audit?".format(
                status=CliUtils.format_color_string(value="ERROR", color="red")))
        parser.print_help()

    def config(self):
        parser = argparse.ArgumentParser(usage="""tj config COMMAND

Allows to configure Test Junkie the way you want it

Commands:
show\t Display current configuration for Test-Junkie
update\t Update configuration settings for individual properties via cli 
restore\t Will restore config to it\'s original values

Use: tj config COMMAND -h to display COMMAND specific help
""")
        parser.add_argument('command', default=None, help="command to run")
        try:
            if len(sys.argv) >= 3:
                command = str(sys.argv[2:3][0])
                if command in ["show", "update", "restore"]:
                    from test_junkie.cli.cli_config import CliConfig
                    return CliConfig(CliConstants.TJ_CONFIG_NAME, command, sys.argv)
                elif command not in ["-h"]:
                    print("[{status}]\t\'{command}\' is not a test-junkie command\n".format(
                          status=CliUtils.format_color_string(value="ERROR", color="red"),
                          command=command))
            else:
                print("[{status}]\tDude, what do you want to do with the config?".format(
                      status=CliUtils.format_color_string(value="ERROR", color="red")))
            parser.print_help()
        except:
            if "SystemExit:" not in traceback.format_exc():
                CliUtils.print_color_traceback()
                parser.print_help()
                exit(120)

    def version(self):
        print("Test Junkie {} (Python{})\n{}".format(pkg_resources.require("test-junkie")[0].version,
                                                     sys.version_info[0],
                                                     DocumentationLinks.DOMAIN))


class CliUtils:

    __INITIALIZED = False

    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    END = '\033[0m'

    def __init__(self):

        pass

    @staticmethod
    def add_standard_tj_args(parser, audit=False):
        """
        Generic parser args used to configure execution or set config settings
        """
        if not audit:
            parser.add_argument("-T", "--test_multithreading_limit", type=int, default=Undefined,
                                help="Test level multi threading allows to run multiple tests concurrently.")

            parser.add_argument("-S", "--suite_multithreading_limit", type=int, default=Undefined,
                                help="Suite level multi threading allows to run multiple suites concurrently.")

            parser.add_argument("-t", "--tests", nargs="+", default=Undefined,
                                help="Test Junkie can run specific tests. "
                                     "Provide the names of the tests that you want to execute/audit.")

        parser.add_argument("-f", "--features", nargs="+", default=Undefined,
                            help="Test suites can be defined with a feature that they are testing. "
                                 "Use features to narrow down execution/audit of test suites only to those that "
                                 "match this filter. Learn more @ {link}".format(link=DocumentationLinks.FEATURES))

        parser.add_argument("-c", "--components", nargs="+", default=Undefined,
                            help="Tests can be defined with a component that they are testing. "
                                 "Use components to narrow down execution/audit of tests only to those that "
                                 "match this filter. Learn more @ {link}".format(link=DocumentationLinks.COMPONENTS))

        parser.add_argument("-o", "--owners", nargs="+", default=Undefined,
                            help="Tests & test suites can be defined with an assignee. "
                                 "Use owners to narrow down execution/audit of tests only to those that "
                                 "match this filter. Learn more @ {link}".format(link=DocumentationLinks.ASSIGNEES))

        if not audit:
            parser.add_argument("-m", "--monitor_resources", action="store_true", default=Undefined,
                                help="Test Junkie can track resource usage for CPU & Memory as it runs tests")

            parser.add_argument("--html_report", type=str, default=Undefined,
                                help="Path to FILE. This will enable HTML report generation and when ready, "
                                     "the report will be saved to the specified file")

            parser.add_argument("--xml_report", type=str, default=Undefined,
                                help="Path to FILE. This will enable XML report generation and when ready, "
                                     "the report will be saved to the specified file")

            parser.add_argument("-l", "--run_on_match_all", nargs="+", default=Undefined,
                                help="Test Junkie will RUN tests that match ALL of the tags. Read more about it: {link}"
                                .format(link=DocumentationLinks.TAGS))

            parser.add_argument("-k", "--run_on_match_any", nargs="+", default=Undefined,
                                help="Test Junkie will RUN tests that match ANY of the tags. Read more about it: {link}"
                                .format(link=DocumentationLinks.TAGS))

            parser.add_argument("-j", "--skip_on_match_all", nargs="+", default=Undefined,
                                help="Test Junkie will SKIP tests that match ALL of the tags. Read more about it: {link}"
                                .format(link=DocumentationLinks.TAGS))

            parser.add_argument("-g", "--skip_on_match_any", nargs="+", default=Undefined,
                                help="Test Junkie will SKIP tests that match ANY of the tags. Read more about it: {link}"
                                .format(link=DocumentationLinks.TAGS))

            parser.add_argument("-q", "--quiet", action="store_true", default=Undefined,
                                help="Suppress all standard output from tests")

            parser.add_argument("--code-cov", action="store_true", default=Undefined,
                                help="Measure code coverage")

            parser.add_argument("--cov-rcfile", type=str, default=Undefined,
                                help="Path to configuration FILE for coverage.py "
                                     "See {link}".format(link=DocumentationLinks.COVERAGE_CONFIG_FILE))
        else:
            parser.add_argument("-l", "--tags", nargs="+", default=Undefined,
                                help="Test Junkie will audit tests that match those tags.")

        parser.add_argument("-s", "--sources", nargs="+", default=Undefined,
                            help="Paths to DIRECTORY or FILE where you have your tests. "
                                 "Test Junkie will traverse this source(s) looking for test suites")

        parser.add_argument("--guess-root", action="store_true", default=Undefined,
                            help="If your project is not part of the PYTHONPATH, you will get an error when running "
                                 "it via command line. If this flag is used, TJ will try to guess the root directory "
                                 "and temporary add it to the path. Usually not recommended.")

    @staticmethod
    def add_standard_boolean_tj_args(parser):
        """
        Generic parser args used to show and restore config settings
        """

        parser.add_argument("-s", "--sources", action="store_true", default=False,
                            help="Paths to DIRECTORY or FILE where you have your tests. "
                                 "Test Junkie will traverse this source(s) looking for test suites")

        parser.add_argument("-T", "--test_multithreading_limit", action="store_true", default=False,
                            help="Test level multi threading allows to run multiple tests concurrently.")

        parser.add_argument("-S", "--suite_multithreading_limit", action="store_true", default=False,
                            help="Suite level multi threading allows to run multiple suites concurrently.")

        parser.add_argument("-t", "--tests", nargs="+", default=Undefined,
                            help="Test Junkie can run specific tests. "
                                 "Provide the names of the tests that you want to run.")

        parser.add_argument("-f", "--features", action="store_true", default=False,
                            help="Test suites can be defined with a feature that they are testing. "
                                 "Use features to narrow down execution of test suites only to those that "
                                 "match this filter. Learn more @ {link}".format(link=DocumentationLinks.FEATURES))

        parser.add_argument("-c", "--components", action="store_true", default=False,
                            help="Tests can be defined with a component that they are testing. "
                                 "Use components to narrow down execution of tests only to those that "
                                 "match this filter. Learn more @ {link}".format(link=DocumentationLinks.COMPONENTS))

        parser.add_argument("-o", "--owners", action="store_true", default=False,
                            help="Tests & test suites can be defined with an assignee. "
                                 "Use owners to narrow down execution of tests only to those that "
                                 "match this filter. Learn more @ {link}".format(link=DocumentationLinks.ASSIGNEES))

        parser.add_argument("-m", "--monitor_resources", action="store_true", default=False,
                            help="Test Junkie can track resource usage for CPU & Memory as it runs tests")

        parser.add_argument("--html_report", action="store_true", default=False,
                            help="Path to FILE. This will enable HTML report generation and when ready, "
                                 "the report will be saved to the specified file")

        parser.add_argument("--xml_report", action="store_true", default=False,
                            help="Path to FILE. This will enable XML report generation and when ready, "
                                 "the report will be saved to the specified file")

        parser.add_argument("-l", "--run_on_match_all", action="store_true", default=False,
                            help="Test Junkie will RUN tests that match ALL of the tags. Read more about it: {link}"
                            .format(link=DocumentationLinks.TAGS))

        parser.add_argument("-k", "--run_on_match_any", action="store_true", default=False,
                            help="Test Junkie will RUN tests that match ANY of the tags. Read more about it: {link}"
                            .format(link=DocumentationLinks.TAGS))

        parser.add_argument("-j", "--skip_on_match_all", action="store_true", default=False,
                            help="Test Junkie will SKIP tests that match ALL of the tags. Read more about it: {link}"
                            .format(link=DocumentationLinks.TAGS))

        parser.add_argument("-g", "--skip_on_match_any", action="store_true", default=False,
                            help="Test Junkie will SKIP tests that match ANY of the tags. Read more about it: {link}"
                            .format(link=DocumentationLinks.TAGS))
        parser.add_argument("-q", "--quiet", action="store_true", default=False,
                            help="Suppress all standard output from tests")
        parser.add_argument("--code-cov", action="store_true", default=False,
                            help="Measure code coverage")
        parser.add_argument("--cov-rcfile", action="store_true", default=False,
                            help="Path to configuration FILE for coverage.py "
                                 "See {link}".format(link=DocumentationLinks.COVERAGE_CONFIG_FILE))
        parser.add_argument("--guess-root", action="store_true", default=False,
                            help="If your project is not part of the PYTHONPATH, you will get an error when running "
                                 "it via command line. If this flag is used, TJ will try to guess the root directory "
                                 "and temporary add it to the path. Usually not recommended.")

    @staticmethod
    def __initialize():
        if not CliUtils.__INITIALIZED:
            import colorama
            colorama.init()
            CliUtils.__INITIALIZED = True

    @staticmethod
    def format_color_string(value, color):
        CliUtils.__initialize()
        colors = {"red": Fore.RED, "green": Fore.GREEN, "yellow": Fore.YELLOW, "blue": Fore.BLUE}
        return "{style}{color}{value}{reset}".format(style=Style.BRIGHT, color=colors[color],
                                                     value=value, reset=Style.RESET_ALL)

    @staticmethod
    def print_color_traceback(trace=None):
        CliUtils.__initialize()
        print(Style.BRIGHT + Fore.RED)
        if trace is None:
            print(traceback.format_exc())
        else:
            print(trace)
        print(Style.RESET_ALL)

    @staticmethod
    def format_bold_string(value):
        CliUtils.__initialize()
        return "{bold}{value}{end}".format(bold=CliUtils.BOLD, value=value, end=CliUtils.END)


if "__main__" == __name__:

    Cli()



import hashlib
import inspect

from test_junkie.constants import DocumentationLinks
from test_junkie.errors import BadParameters, BadSignature
from test_junkie.listener import Listener
from test_junkie.rules import Rules


class Builder(object):

    __TEST_ID = 0  # unique but not persistent from execution to execution
    __SUITE_ID = 0  # unique but not persistent from execution to execution

    __UNIQUE_TEST_SUITES = []
    __EXECUTION_ROSTER = {}
    __CURRENT_SUITE_OBJECT = None
    __SUITE_VALIDATION_ARGS = {"owner": [str], "meta": [dict], "retry": [int], "listener": [Listener], "rules": [Rules],
                               "parallelized": [bool], "priority": [int], "feature": [str], "pr": [list],
                               "parameters": ["<type 'function'>", list], "skip": ["<type 'function'>", bool]}
    __TEST_VALIDATION_ARGS = {"owner": [str], "meta": [dict], "retry": [int], "parallelized_parameters": [bool],
                              "parallelized": [bool], "priority": [int], "component": [str], "tags": [list],
                              "no_retry_on": [list], "retry_on": [list], "pr": [list],
                              "parameters": ["<type 'function'>", list], "skip": ["<type 'function'>", bool]}

    __GROUP_RULES = []
    __GROUP_RULE_DEFINITIONS = {}
    __REQUESTED_SUITES = None

    @staticmethod
    def get_execution_roster():
        return Builder.__EXECUTION_ROSTER

    @staticmethod
    def __set_current_suite_object_defaults():
        from test_junkie.decorators import DecoratorType
        Builder.__CURRENT_SUITE_OBJECT = {"test_listener": None,
                                          "test_rules": None,
                                          "class_name": None,
                                          "class_retry": None,
                                          "class_skip": False,
                                          "class_object": None,
                                          "suite_definition": {DecoratorType.BEFORE_CLASS: [],
                                                               DecoratorType.BEFORE_TEST: [],
                                                               DecoratorType.TEST_CASE: [],
                                                               DecoratorType.AFTER_TEST: [],
                                                               DecoratorType.AFTER_CLASS: []}}

    @staticmethod
    def build_group_definitions(suites):

        Builder.__REQUESTED_SUITES = suites
        for group_rule in Builder.__GROUP_RULES:
            func = group_rule["decorated_function"]
            func(func)
        from test_junkie.objects import GroupRulesObject
        return GroupRulesObject(Builder.__GROUP_RULE_DEFINITIONS)

    @staticmethod
    def register_group_rules(decorated_function, decorator_kwargs, decorator_type):

        Builder.__GROUP_RULES.append({"decorated_function": decorated_function,
                                      "decorator_kwargs": decorator_kwargs,
                                      "decorator_type": decorator_type})

    @staticmethod
    def add_group_rule(suites, decorated_function, decorator_kwargs, decorator_type):

        if not suites or not isinstance(suites, list):
            raise BadParameters("Group Rules must be defined with a mandatory argument \"suites\" which must be "
                                "of type {}. Please see documentation: {}".format(list, DocumentationLinks.GROUP_RULES))
        suites = sorted(set(suites), key=lambda x: str(x), reverse=True)
        for suite in list(suites):
            if suite not in Builder.__REQUESTED_SUITES:
                suites.remove(suite)
        if suites:  # making sure that rules only apply when we actually run applicable test suites
            group = hashlib.md5(str(suites).encode("utf8")).hexdigest()
            if group not in Builder.__GROUP_RULE_DEFINITIONS:
                Builder.__GROUP_RULE_DEFINITIONS.update({group: {"suites": suites, "rules": {}}})
            if decorator_type not in Builder.__GROUP_RULE_DEFINITIONS[group]["rules"]:
                Builder.__GROUP_RULE_DEFINITIONS[group]["rules"].update({decorator_type: []})
            Builder.__GROUP_RULE_DEFINITIONS[group]["rules"][decorator_type].append(
                {"decorated_function": decorated_function, "decorator_kwargs": decorator_kwargs})

    @staticmethod
    def build_suite_definitions(decorated_function, decorator_kwargs, decorator_type):
        from test_junkie.decorators import DecoratorType
        from test_junkie.debugger import LogJunkie
        from test_junkie.objects import SuiteObject

        if decorator_type == DecoratorType.TEST_CASE:
            Builder.__TEST_ID += 1
        elif decorator_type == DecoratorType.TEST_SUITE:
            Builder.__SUITE_ID += 1

        _function_name = None
        _class_name = None
        if inspect.isfunction(decorated_function):
            decorator_kwargs.update({"testjunkie_test_id": Builder.__TEST_ID,
                                     "testjunkie_suite_id": Builder.__SUITE_ID})
            Builder.__validate_test_kwargs(decorator_kwargs, decorated_function)
            _function_name = decorated_function.__name__
        else:
            if decorated_function is not None:
                Builder.__validate_suite_kwargs(decorator_kwargs)
                _class_name = decorated_function.__name__
                Builder.__CURRENT_SUITE_OBJECT["class_object"] = decorated_function
                Builder.__CURRENT_SUITE_OBJECT["class_retry"] = decorator_kwargs.get("retry", 1)
                Builder.__CURRENT_SUITE_OBJECT["class_skip"] = decorator_kwargs.get("skip", False)
                Builder.__CURRENT_SUITE_OBJECT["class_meta"] = decorator_kwargs.get("meta", {})
                Builder.__CURRENT_SUITE_OBJECT["test_listener"] = decorator_kwargs.get("listener", Listener)
                Builder.__CURRENT_SUITE_OBJECT["test_rules"] = decorator_kwargs.get("rules", Rules)
                Builder.__CURRENT_SUITE_OBJECT["class_parameters"] = decorator_kwargs.get("parameters", [None])
                Builder.__CURRENT_SUITE_OBJECT["parallelized"] = decorator_kwargs.get("parallelized", True)
                decorator_kwargs.update({"testjunkie_suite_id": Builder.__SUITE_ID})
                Builder.__CURRENT_SUITE_OBJECT["decorator_kwargs"] = decorator_kwargs

        if Builder.__CURRENT_SUITE_OBJECT is None:
            Builder.__set_current_suite_object_defaults()
        elif Builder.__CURRENT_SUITE_OBJECT.get("class_name", None) is None:
            Builder.__CURRENT_SUITE_OBJECT["class_name"] = _class_name

        if _function_name is not None:
            Builder.__CURRENT_SUITE_OBJECT["suite_definition"][decorator_type].append(
                {"decorated_function": decorated_function, "decorator_kwargs": decorator_kwargs})
            LogJunkie.debug("=======================Suite Definition Updated=============================")
            LogJunkie.debug("Function: {}".format(_function_name))
            LogJunkie.debug("Decorator Type: {}".format(decorator_type))
            LogJunkie.debug("Decorator Arguments: {}".format(decorator_kwargs))
            LogJunkie.debug("Function object: {}".format(decorated_function))
            LogJunkie.debug("============================================================================")
        else:
            LogJunkie.debug("=======================Suite Definition Finished=============================")
            LogJunkie.debug("Suite: {}".format(_class_name))
            LogJunkie.debug("Suite Definition: {}".format(Builder.__CURRENT_SUITE_OBJECT))
            Builder.__EXECUTION_ROSTER.update({decorated_function: SuiteObject(Builder.__CURRENT_SUITE_OBJECT)})
            Builder.__set_current_suite_object_defaults()
            LogJunkie.debug(">> Definition reset for next suite <<")
            LogJunkie.debug("=============================================================================\n\n")

    # noinspection PyTypeHints
    @staticmethod
    def __validation_failed(kwargs, suite=True):

        def validation_failed(actual, expected):
            if expected == "<type 'function'>":
                if inspect.isfunction(actual) or inspect.ismethod(actual):
                    return False
            else:
                if inspect.isclass(actual) and issubclass(actual, expected):
                    return False
                if type(actual) is expected or isinstance(actual, expected):
                    return False
            return True

        args = Builder.__SUITE_VALIDATION_ARGS if suite else Builder.__TEST_VALIDATION_ARGS
        for arg, expected_types in args.items():
            failed = True
            if arg in kwargs:
                for expected_type in expected_types:
                    if failed:
                        failed = validation_failed(kwargs.get(arg), expected_type)
                if failed:
                    return {"expected": expected_types, "actual": type(kwargs.get(arg)), "arg": arg}
        return False

    @staticmethod
    def __validate_suite_kwargs(kwargs):

        data = Builder.__validation_failed(kwargs)
        if data:
            raise BadParameters("Argument: \"{}\" in @Suite() decorator must be of either type: {} but found: {}. "
                                "For more info, see @Suite() decorator documentation: {}"
                                .format(data["arg"], data["expected"], data["actual"],
                                        DocumentationLinks.SUITE_DECORATOR))

    @staticmethod
    def __validate_test_kwargs(kwargs, decorated_function):
        data = Builder.__validation_failed(kwargs, suite=False)
        if data:
            raise BadParameters("Argument: \"{}\" in @test() decorator must be of either type: {} but found: {}. "
                                "For more info, see @test() decorator documentation: {}"
                                .format(data["arg"], data["expected"], data["actual"],
                                        DocumentationLinks.TEST_DECORATOR))
        if "parameter" not in inspect.getargspec(decorated_function).args and kwargs.get("parameters") is not None:
            raise BadSignature("When using \"parameters\" argument for @test() decorator, "
                               "you must accept \"parameter\" in the function's signature. "
                               "For more info, see documentation: {}"
                               .format(DocumentationLinks.PARAMETERIZED_TESTS))



import pprint
from test_junkie.runner import Runner
from tests.QualityManager import QualityManager
from tests.junkie_suites.AfterClassExceptionSuite import AfterClassExceptionSuite

runner = Runner([AfterClassExceptionSuite])
runner.run()
results = runner.get_executed_suites()

pprint.pprint(results[0].metrics.get_metrics())
for test in results[0].get_test_objects():
    print(test.get_function_name())
    pprint.pprint(test.metrics.get_metrics())


def test_class_metrics():

    metrics = results[0].metrics.get_metrics()
    QualityManager.check_class_metrics(metrics,
                                       expected_beforeclass_exception_count=1,
                                       expected_beforeclass_exception_object=None,
                                       expected_beforeclass_performance_count=1,

                                       expected_afterclass_exception_count=1,
                                       expected_afterclass_exception_object=Exception,
                                       expected_afterclass_performance_count=1)


def test_test_metrics():

    assert results[0].get_test_objects()
    for test in results[0].get_test_objects():

        metrics = test.metrics.get_metrics()["None"]["None"]
        QualityManager.check_test_metrics(metrics)



import pprint

from test_junkie.constants import TestCategory
from test_junkie.runner import Runner
from tests.QualityManager import QualityManager
from tests.junkie_suites.FeatureAggregations import LoginSessions, Login, Dashboard

runner = Runner([Login, LoginSessions, Dashboard])
runner_metrics = runner.run()
results = runner.get_executed_suites()
tests = results[0].get_test_objects()
pprint.pprint(results[0].metrics.get_metrics())
for test in results[0].get_test_objects():
    print(test.get_function_name())
    pprint.pprint(test.metrics.get_metrics())


def test_class_metrics():

    metrics = results[0].metrics.get_metrics()
    QualityManager.check_class_metrics(metrics)


def test_test_metrics():
    for test in tests:
        metrics = test.metrics.get_metrics()["None"]["None"]
        QualityManager.check_test_metrics(metrics)


def test_advanced_aggregation_metrics():
    metrics = runner_metrics.get_report_by_features()
    dashboard_report = metrics["Dashboard"]["Charts"]
    login_auth_report = metrics["Login"]["Authentication"]
    login_input_report = metrics["Login"]["Login Inputs"]
    login_session_report = metrics["Login"]["Session Timeout"]
    reports = [dashboard_report, login_auth_report, login_input_report, login_session_report]
    for report in reports:
        assert report[TestCategory.CANCEL] == 0
        assert report[TestCategory.ERROR] == 0
        assert len(report["exceptions"]) == 0
        assert len(report["performance"]) == 2
        assert report[TestCategory.FAIL] == 0
        assert report[TestCategory.IGNORE] == 0
        assert report[TestCategory.SKIP] == 0
        assert report["retries"] == [1, 1]
        assert report["total"] == 2


def test_basic_aggregation_metrics():
    metrics = runner_metrics.get_basic_report()["tests"]
    assert metrics["total"] == 8
    assert metrics[TestCategory.SUCCESS] == 8
    assert metrics[TestCategory.FAIL] == 0
    assert metrics[TestCategory.ERROR] == 0
    assert metrics[TestCategory.IGNORE] == 0
    assert metrics[TestCategory.SKIP] == 0
    assert metrics[TestCategory.CANCEL] == 0



# import datetime
# import time
# from random import randint
#
# from playground.suiteB import ShoppingCartSuite
# from test_junkie.decorators import Suite, test, beforeTest, beforeClass, afterTest
# from test_junkie.meta import Meta
# from tests.junkie_suites.Constants import Constants
# from tests.junkie_suites.TestListener import TestListener
# from tests.junkie_suites.TestRules import TestRules
#
#
# def test_func():
#     print("evaluating...")
#     # time.sleep(2)
#     return [1, 2]
#
#
# @Suite(feature="Login", owner="Mike", rules=TestRules, listener=TestListener)
# class LoginSuite:
#
#     # @beforeClass()
#     # def before_class(self):
#     #     time.sleep(randint(0, 4))
#     #
#     # @afterTest()
#     # def after_test(self):
#     #     time.sleep(randint(0, 4))
#     #     raise Exception("Winning!")
#
#     # @beforeTest()
#     # def before_test(self):
#     #     time.sleep(randint(0, 4))
#     #     raise Exception("Winning!")
#
#     @test(component="Login Page",
#           tags=["positive_flow", "ui", "auth"], parameters=[1, 2])
#     def positive_login(self, parameter):
#         pass
#         # time.sleep(randint(0, 4))
#
#     # @test(priority=1, skip_before_test=True,
#     #       parallelized_parameters=True,
#     #       component="Login Page",
#     #       tags=["negative_flow", "ui"],
#     #       parameters=[{"pass": ""}, {"pass": "example"}], retry=3)
#     # def negative_login(self, parameter, suite_parameter):
#     #     # time.sleep(1)
#     #     time.sleep(randint(0, 4))
#     #     raise AssertionError("Missing error message on negative login attempt: {}".format(parameter))
#     #
#     # @test(component="Session", owner="Victor", skip_after_test=True,
#     #       tags=["positive_flow", "ui", "auth", "session"])
#     # def session_timeout_after_2h(self):
#     #     pass
#     #
#     # @test(component="Session", owner="Victor",
#     #       tags=["positive_flow", "ui", "auth", "session"])
#     # def session_timeout_after_1h(self):
#     #     pass
#     #
#     # @test(component="Session", owner="Victor",
#     #       tags=["negative_flow", "ui", "auth", "session"])
#     # def logout_after_session_timeout(self):
#     #     print("Running a.a")
#     #     print(datetime.datetime.now())
#     #     # print(time.sleep(7.3))
#     #     time.sleep(randint(0, 7))
#     #
#     # @test(component="Login Page", skip_before_test=True,
#     #       tags=["negative_flow", "ui"], pr=[ShoppingCartSuite.add_to_cart])
#     # def negative_login_attempt_limit(self):
#     #     print("Running a")
#     #     print(datetime.datetime.now())
#     #     # print(time.sleep(7.3))
#     #     time.sleep(randint(0, 7))
#     #     pass



# from test_junkie.decorators import beforeClass, afterClass, afterTest, beforeTest

#
# class TestRules:
#
#     @beforeClass()
#     def before_class(self):
#         pass
#
#     @beforeTest()
#     def before_test(self):
#         # write your code here
#         pass
#
#     @afterTest()
#     def after_test(self):
#         # write your code here
#         pass
#
#     @afterClass()
#     def after_class(self):
#         # write your code here
#         pass
from playground.suiteA import LoginSuite

GROUP = [LoginSuite]



import os
import pprint

from test_junkie.constants import CliConstants
from test_junkie.runner import Runner
from test_junkie.cli.cli_config import Config
from tests.QualityManager import QualityManager
from tests.cli.CliTestSuite import AuthApiSuite, ShoppingCartSuite, NewProductsSuite
from tests.cli.Cmd import Cmd

ROOT = __file__.split("{0}tests".format(os.sep))[0]
EXE = ROOT + "{0}test_junkie{0}cli{0}cli.py".format(os.sep)
TESTS = ROOT + "{0}tests{0}cli".format(os.sep)


def test_help():

    commands = [['python', EXE, '-h'],
                ['python', EXE, 'run', '-h'],
                ['python', EXE, 'audit', '-h'],
                ['python', EXE, 'config', '-h'],
                ['python', EXE, 'version', '-h']]
    for cmd in commands:
        output = Cmd.run(cmd)
        for line in output:
            assert "Traceback (most recent call last)" not in line, \
                "Command: {} produced exception".format(cmd, output)


def test_sub_command_help():

    commands = [['python', EXE, 'audit', 'features', '-h'],
                ['python', EXE, 'audit', 'tags', '-h'],
                ['python', EXE, 'audit', 'components', '-h'],
                ['python', EXE, 'audit', 'owners', '-h'],
                ['python', EXE, 'audit', 'suites', '-h'],
                ['python', EXE, 'config', 'update', '-h'],
                ['python', EXE, 'config', 'show', '-h'],
                ['python', EXE, 'config', 'restore', '-h']]
    for cmd in commands:
        output = Cmd.run(cmd)
        for line in output:
            assert "Traceback (most recent call last)" not in line, \
                "Command: {} produced exception. {}".format(cmd, output)


def test_incomplete_inputs():

    commands = [
                ['python', EXE, 'config'],
                ['python', EXE, 'config', 'update'],
                ['python', EXE, 'config', 'restore'],
                ['python', EXE, 'config', 'show'],
                ['python', EXE, 'audit'],
                ['python', EXE, 'version']]
    for cmd in commands:
        output = Cmd.run(cmd)
        for line in output:
            assert "Traceback (most recent call last)" not in line, \
                "Command: {} produced exception. {}".format(cmd, output)


def test_config_update():

    commands = [['python', EXE, 'config', 'update', '--test_multithreading_limit', '10'],
                ['python', EXE, 'config', 'update', '--suite_multithreading_limit', '10'],
                ['python', EXE, 'config', 'update', '--features', 'feat'],
                ['python', EXE, 'config', 'update', '--components', 'comp'],
                ['python', EXE, 'config', 'update', '--owners', 'own'],
                ['python', EXE, 'config', 'update', '--monitor_resources'],
                ['python', EXE, 'config', 'update', '--html_report', '/test/path/for/html'],
                ['python', EXE, 'config', 'update', '--xml_report', '/test/path/for/xml'],
                ['python', EXE, 'config', 'update', '--run_on_match_all', '1000'],
                ['python', EXE, 'config', 'update', '--run_on_match_any', '2000'],
                ['python', EXE, 'config', 'update', '--skip_on_match_all', '3000'],
                ['python', EXE, 'config', 'update', '--skip_on_match_any', '4000'],
                ]
    for cmd in commands:
        output = Cmd.run(cmd)

        prop = cmd[4].replace("--", "")
        value = cmd[5] if len(cmd) == 6 else "True"
        assert "Traceback (most recent call last)" not in output[-2], \
            "Command: {} produced exception. {}".format(cmd, output)
        assert "OK" in output[-2]
        assert prop in output[-2], "Command: {} did not update property: {}".format(cmd, prop)
        assert value in output[-2], "Command: {} did not update property: {} to value: {}".format(cmd, prop, value)


def test_config_restore():

    commands = [['python', EXE, 'config', 'restore', '--test_multithreading_limit'],
                ['python', EXE, 'config', 'restore', '--suite_multithreading_limit'],
                ['python', EXE, 'config', 'restore', '--features'],
                ['python', EXE, 'config', 'restore', '--components'],
                ['python', EXE, 'config', 'restore', '--owners'],
                ['python', EXE, 'config', 'restore', '--monitor_resources'],
                ['python', EXE, 'config', 'restore', '--html_report'],
                ['python', EXE, 'config', 'restore', '--xml_report'],
                ['python', EXE, 'config', 'restore', '--run_on_match_all'],
                ['python', EXE, 'config', 'restore', '--run_on_match_any'],
                ['python', EXE, 'config', 'restore', '--skip_on_match_all'],
                ['python', EXE, 'config', 'restore', '--skip_on_match_any'],
                ]
    for cmd in commands:
        output = Cmd.run(cmd)
        prop = cmd[4].replace("--", "")
        value = "None"
        for line in output:
            assert "Traceback (most recent call last)" not in line, \
                "Command: {} produced exception. {}".format(cmd, output)
        assert "OK" in output[0]
        assert prop in output[0], "Command: {} did not update property: {}".format(cmd, prop)
        assert value in output[0], "Command: {} did not update property: {} to value: {}".format(cmd, prop, value)


def test_audit_by_owner():

    Cmd.run(['python', EXE, 'config', 'restore', '--all'])
    commands = [['python', EXE, 'audit', 'owners', '-s', TESTS],
                ['python', EXE, 'audit', 'owners', '-s', TESTS, '-o', 'Mike']]
    for cmd in commands:
        output = Cmd.run(cmd)
        assert_not_in = ["Tag:", "Component:", "Owners:", "Feature:", "Suite:"]
        if "Mike" in cmd:
            assert_not_in.append("Owner: George")
            assert_not_in.append("Owner: Victor")
            assert_not_in.append("None")
        not_validated = ["Owner:", "Tests:", "Components:", "Tags:", "Features:", "Suites:"]
        for line in output:
            assert "Traceback (most recent call last)" not in line, \
                "Command: {} produced exception. {}".format(cmd, output)
            assert "ERROR" not in line
            for item in assert_not_in:
                assert item not in line
            for validate in list(not_validated):
                if validate in line:
                    not_validated.remove(validate)
        assert len(not_validated) == 0


def test_audit_by_suite():

    Cmd.run(['python', EXE, 'config', 'restore', '--all'])
    commands = [['python', EXE, 'audit', 'suites', '-s', TESTS],
                ['python', EXE, 'audit', 'suites', '-s', TESTS, '-o', 'Mike']]
    for cmd in commands:
        output = Cmd.run(cmd)
        assert_not_in = ["Tag:", "Owner:", "Component:", "Suites:"]
        if "Mike" in cmd:
            assert_not_in.append("Victor")
            assert_not_in.append("George")
            assert_not_in.append("None")
            assert_not_in.append("Suites:")
        not_validated = ["Mike", "Owners:", "Feature:", "Components:", "Tags:", "Suite:"]
        for line in output:
            assert "Traceback (most recent call last)" not in line, \
                "Command: {} produced exception. {}".format(cmd, output)
            assert "ERROR" not in line
            for item in assert_not_in:
                assert item not in line
            for validate in list(not_validated):
                if validate in line:
                    not_validated.remove(validate)
        assert len(not_validated) == 0


def test_audit_by_tags():
    Cmd.run(['python', EXE, 'config', 'restore', '--all'])
    commands = [['python', EXE, 'audit', 'tags', '-s', TESTS],
                ['python', EXE, 'audit', 'tags', '-s', TESTS, '-l', 'sso']]
    for cmd in commands:
        output = Cmd.run(cmd)
        assert_not_in = ["Tags:", "Owner:", "Component:", "Suite:", "Feature:"]
        not_validated = ["Tag:", "Mike", "Owners:", "Features:", "Suites:", "Components:", "API (2)", "2 of 16"]
        for line in output:
            assert "Traceback (most recent call last)" not in line, \
                "Command: {} produced exception. {}".format(cmd, output)
            assert "ERROR" not in line
            for item in assert_not_in:
                assert item not in line
            for validate in list(not_validated):
                if validate in line:
                    not_validated.remove(validate)
        assert len(not_validated) == 0


def test_audit_by_features():
    Cmd.run(['python', EXE, 'config', 'restore', '--all'])
    commands = [['python', EXE, 'audit', 'features', '-s', TESTS],
                ['python', EXE, 'audit', 'features', '-s', TESTS, '-f', 'Store']]
    for cmd in commands:
        output = Cmd.run(cmd)
        assert_not_in = ["Tag:", "Owner:", "Component:", "Suite:", "Features:"]
        not_validated = ["Feature:", "Mike (5)", "George (5)", "Owners:", "Tags:", "Suites:", "Components:"]
        for line in output:
            assert "Traceback (most recent call last)" not in line, \
                "Command: {} produced exception. {}".format(cmd, output)
            assert "ERROR" not in line
            for item in assert_not_in:
                assert item not in line
            for validate in list(not_validated):
                if validate in line:
                    not_validated.remove(validate)
        assert len(not_validated) == 0


def test_audit_by_components():
    Cmd.run(['python', EXE, 'config', 'restore', '--all'])
    commands = [['python', EXE, 'audit', 'components', '-s', TESTS],
                ['python', EXE, 'audit', 'components', '-s', TESTS, '-c', 'Admin']]
    for cmd in commands:
        output = Cmd.run(cmd)
        assert_not_in = ["Tag:", "Owner:", "Components:", "Suite:", "Feature:"]
        not_validated = ["Component:", "Mike (5)", "Owners:", "Tags:", "Suites:", "Features:"]
        for line in output:
            assert "Traceback (most recent call last)" not in line, \
                "Command: {} produced exception. {}".format(cmd, output)
            assert "ERROR" not in line
            for item in assert_not_in:
                assert item not in line
            for validate in list(not_validated):
                if validate in line:
                    not_validated.remove(validate)
        assert len(not_validated) == 0


def test_bad_inputs():

    commands = [['python', EXE, 'run'],
                ['python', EXE, 'audit'],
                ['python', EXE, 'config'],
                ['python', EXE, 'configgg'],
                ['python', EXE, 'config', 'restore'],
                ['python', EXE, 'config', 'update'],
                ['python', EXE, 'config', 'updateee']]
    for cmd in commands:
        output = Cmd.run(cmd)
        for line in output:
            assert "Traceback (most recent call last)" not in line, \
                "Command: {} produced exception. {}".format(cmd, output)
        validate_output(expected=[["ERROR", ""]], output=output)


def test_config_restore_all():

    commands = [['python', EXE, 'config', 'restore', '--all']]
    for cmd in commands:
        output = Cmd.run(cmd)
        for line in output:
            assert "Traceback (most recent call last)" not in line, \
                "Command: {} produced exception. {}".format(cmd, output)
        assert "Config restored to default settings!" in output[-1], "Wrong message: {}".format(output[-1])


def test_config_show_all():

    commands = [['python', EXE, 'config', 'show', '--all']]
    for cmd in commands:
        output = Cmd.run(cmd)
        for line in output:
            assert "Traceback (most recent call last)" not in line, \
                "Command: {} produced exception. {}".format(cmd, output)


def validate_output(expected, output):

    for item in list(expected):
        for line in output:
            if item[0] in line and item[1] in line:
                expected.remove(item)
    if expected:
        raise AssertionError("Not verified: {}".format(expected))


def test_run_with_cmd_args():

    Cmd.run(['python', EXE, 'config', 'restore', '--all'])
    for cmd in [['python', EXE, 'run', '-s', TESTS, '-k', 'api'],
                ['python', EXE, 'run', '-s', TESTS, '-k', 'api', '-q'],
                ['python', EXE, 'run', '-s', TESTS, '-k', 'api', '--code-cov'],
                ['python', EXE, 'run', '-s', TESTS, '-k', 'api', '--code-cov', '-q']]:
        output = Cmd.run(cmd)
        pprint.pprint(output)
        validate_output(expected=[["[6/16 37.50%]", "SUCCESS"]], output=output)


def test_run_with_config_and_cmd_args():

    Cmd.run(['python', EXE, 'config', 'restore', '--all'])
    Cmd.run(['python', EXE, 'config', 'update', '-k', 'api ui'])
    Cmd.run(['python', EXE, 'config', 'update', '-g', 'sso'])
    output = Cmd.run(['python', EXE, 'run', '-s', TESTS, '-k', 'api'])
    pprint.pprint(output)
    validate_output(expected=[["[4/16 25.00%]", "SUCCESS"]], output=output)


def test_run_with_config():

    Cmd.run(['python', EXE, 'config', 'restore', '--all'])
    Cmd.run(['python', EXE, 'config', 'update', '-k', 'api', 'ui'])
    Cmd.run(['python', EXE, 'config', 'update', '-g', 'sso'])
    output = Cmd.run(['python', EXE, 'run', '-s', TESTS])
    pprint.pprint(output)
    validate_output(expected=[["[9/16 56.25%]", "SUCCESS"]], output=output)


def test_runner_with_config():

    Cmd.run(['python', EXE, 'config', 'restore', '--all'])
    Cmd.run(['python', EXE, 'config', 'update', '-k', 'ui'])
    Cmd.run(['python', EXE, 'config', 'update', '-g', 'sso'])
    runner = Runner(suites=[ShoppingCartSuite, AuthApiSuite],
                    config=Config.get_config_path(CliConstants.TJ_CONFIG_NAME))
    runner.run()

    results = runner.get_executed_suites()
    tests_ShoppingCartSuite = results[0].get_test_objects()
    tests_AuthApiSuite = results[1].get_test_objects()
    pprint.pprint(results[0].metrics.get_metrics())

    metrics = results[0].metrics.get_metrics()
    QualityManager.check_class_metrics(metrics, expected_status="success")
    print(len(tests_ShoppingCartSuite))
    assert len(tests_ShoppingCartSuite) == 5, \
        "Expected 5 tests to be executed, Actually executed: {}".format(len(tests_ShoppingCartSuite))
    assert len(tests_AuthApiSuite) == 6, \
        "Expected 6 tests to be executed, Actually executed: {}".format(len(tests_AuthApiSuite))
    for test in tests_ShoppingCartSuite:
        for class_param, class_data in test.metrics.get_metrics().items():
            for param, metrics in class_data.items():
                QualityManager.check_test_metrics(metrics,
                                                  expected_status="success",
                                                  expected_param=param,
                                                  expected_class_param=metrics["class_param"])
    for test in tests_AuthApiSuite:
        for class_param, class_data in test.metrics.get_metrics().items():
            for param, metrics in class_data.items():
                QualityManager.check_test_metrics(metrics,
                                                  expected_status="skip",
                                                  expected_param=param,
                                                  expected_class_param=metrics["class_param"],
                                                  expected_retry_count=2,
                                                  expected_exception_count=2,
                                                  expected_performance_count=2)
                # should be 1 but there is a bug fix for which was reverted so will revisit once fixed


def test_runner_without_config():

    Cmd.run(['python', EXE, 'config', 'restore', '--all'])
    Cmd.run(['python', EXE, 'config', 'update', '-k', 'ui'])
    Cmd.run(['python', EXE, 'config', 'update', '-g', 'sso'])

    runner = Runner(suites=[NewProductsSuite])
    runner.run()
    results = runner.get_executed_suites()
    tests = results[0].get_test_objects()
    metrics = results[0].metrics.get_metrics()
    QualityManager.check_class_metrics(metrics, expected_status="success")

    assert len(tests) == 5, "Expected 5 tests to be executed, Actually executed: {}".format(len(tests))

    for test in tests:
        for class_param, class_data in test.metrics.get_metrics().items():
            for param, metrics in class_data.items():
                QualityManager.check_test_metrics(metrics,
                                                  expected_status="success",
                                                  expected_param=param,
                                                  expected_class_param=metrics["class_param"])



import pprint
import threading
import sys
sys.path.insert(1, __file__.split("tests")[0])
from tests.QualityManager import QualityManager
from test_junkie.runner import Runner
from tests.junkie_suites.CancelSuite import CancelSuite

runner = Runner([CancelSuite])
thread = threading.Thread(target=runner.run, args=())
runner.cancel()
thread.start()
thread.join()
results = runner.get_executed_suites()

pprint.pprint(results[0].metrics.get_metrics())
for test in results[0].get_test_objects():
    print(test.get_function_name())
    pprint.pprint(test.metrics.get_metrics())


def test_class_metrics():

    metrics = results[0].metrics.get_metrics()
    QualityManager.check_class_metrics(metrics,
                                       expected_retry_count=0,
                                       expected_status="cancel")


def test_test_metrics():

    assert results[0].get_test_objects()
    for test in results[0].get_test_objects():

        properties = test.metrics.get_metrics()
        assert len(properties) == 0





class QualityManager:

    @staticmethod
    def check_class_metrics(metrics,
                            expected_retry_count=1,
                            expected_status="success",
                            expected_runtime=0,

                            expected_afterclass_exception_count=0,
                            expected_beforeclass_exception_count=0,
                            expected_aftertest_exception_count=0,
                            expected_beforetest_exception_count=0,

                            expected_afterclass_exception_object=None,
                            expected_beforeclass_exception_object=None,
                            expected_aftertest_exception_object=None,
                            expected_beforetest_exception_object=None,

                            expected_afterclass_performance_count=0,
                            expected_beforeclass_performance_count=0,
                            expected_aftertest_performance_count=0,
                            expected_beforetest_performance_count=0,

                            expected_afterclass_performance_time=0,
                            expected_beforeclass_performance_time=0,
                            expected_aftertest_performance_time=0,
                            expected_beforetest_performance_time=0):

        assert metrics["retry"] == expected_retry_count, \
            "Expected retry count: {} Actual retry count: {}".format(expected_retry_count, metrics["retry"])
        assert metrics["status"] == expected_status
        assert metrics["runtime"] >= expected_runtime

        assert len(metrics["afterClass"]["exceptions"]) == expected_afterclass_exception_count
        for i in metrics["afterClass"]["exceptions"]:
            assert type(i) == type(expected_afterclass_exception_object) \
                if not isinstance(expected_afterclass_exception_object, type) else expected_afterclass_exception_object

        assert len(metrics["afterClass"]["performance"]) == expected_afterclass_performance_count
        for i in metrics["afterClass"]["performance"]:
            assert i >= expected_afterclass_performance_time

        assert len(metrics["beforeClass"]["exceptions"]) == expected_beforeclass_exception_count
        for i in metrics["beforeClass"]["exceptions"]:
            assert type(i) == type(expected_beforeclass_exception_object) \
                if not isinstance(expected_beforeclass_exception_object, type) else expected_beforeclass_exception_object

        assert len(metrics["beforeClass"]["performance"]) == expected_beforeclass_performance_count
        for i in metrics["beforeClass"]["performance"]:
            assert i >= expected_beforeclass_performance_time

        assert len(metrics["afterTest"]["exceptions"]) == expected_aftertest_exception_count, \
            "Expected: {} Actual: {}".format(expected_aftertest_exception_count,
                                             len(metrics["afterTest"]["exceptions"]))
        for i in metrics["afterTest"]["exceptions"]:
            assert type(i) == type(expected_aftertest_exception_object) \
                if not isinstance(expected_aftertest_exception_object, type) else expected_aftertest_exception_object

        assert len(metrics["afterTest"]["performance"]) == expected_aftertest_performance_count, \
            "Expected: {} Actual: {}".format(expected_aftertest_performance_count,
                                             len(metrics["afterTest"]["performance"]))
        for i in metrics["afterTest"]["performance"]:
            assert i >= expected_aftertest_performance_time

        assert len(metrics["beforeTest"]["exceptions"]) == expected_beforetest_exception_count, \
            "Expected: {} Actual: {}".format(expected_beforetest_exception_count,
                                             len(metrics["beforeTest"]["exceptions"]))
        for i in metrics["beforeTest"]["exceptions"]:
            assert type(i) == type(expected_beforetest_exception_object) \
                if not isinstance(expected_beforetest_exception_object, type) else expected_beforetest_exception_object

        assert len(metrics["beforeTest"]["performance"]) == expected_beforetest_performance_count, \
            "Expected: {} Actual: {}".format(expected_beforetest_performance_count,
                                             len(metrics["beforeTest"]["performance"]))
        for i in metrics["beforeTest"]["performance"]:
            assert i >= expected_beforetest_performance_time

    @staticmethod
    def check_test_metrics(metrics,
                           expected_retry_count=1,
                           expected_status="success",
                           expected_param=None,
                           expected_class_param=None,
                           expected_exception_count=1,
                           expected_exception=None,
                           expected_performance_count=1,
                           expected_performance=0):

        assert metrics["status"] == expected_status, \
            "Expected status: {} Actual Status: {}".format(expected_status, metrics["status"])
        assert metrics["retry"] == expected_retry_count
        assert str(metrics["param"]) == str(expected_param)
        assert str(metrics["class_param"]) == str(expected_class_param)

        assert len(metrics["exceptions"]) == expected_exception_count
        for i in metrics["exceptions"]:
            assert type(i) == type(expected_exception) \
                if not isinstance(expected_exception, type) else expected_exception

        assert len(metrics["performance"]) == expected_performance_count
        for i in metrics["performance"]:
            assert i >= expected_performance



from test_junkie.constants import TestCategory
from test_junkie.runner import Runner
from tests.QualityManager import QualityManager
from tests.junkie_suites.SkipSuites import SkipOwner


def test_component():

    runner = Runner([SkipOwner])
    aggregator = runner.run(owners=["Jane Doe"])
    metrics = aggregator.get_basic_report()["tests"]

    assert metrics[TestCategory.SUCCESS] == 1
    assert metrics[TestCategory.SKIP] == 2

    suites = runner.get_executed_suites()
    metrics = suites[0].metrics.get_metrics()
    QualityManager.check_class_metrics(metrics, expected_status="success")



from test_junkie.decorators import Suite, test
from tests.junkie_suites.TestListener import TestListener
from tests.junkie_suites.error_handling.BadAfterTestRules import BadAfterTestRules


@Suite(listener=TestListener, rules=BadAfterTestRules)
class ErrorSuite5:

    @test()
    def test_1(self, parameter):
        pass
